import hashlib
import pathlib
from abc import ABC, abstractmethod
from collections.abc import Iterable
from typing import NamedTuple, Optional, cast

from competitive_verifier import config


class TestCaseFile(NamedTuple):
    name: str
    input_path: pathlib.Path
    output_path: pathlib.Path


class TestCaseData(NamedTuple):
    name: str
    input_data: bytes
    output_data: bytes


class TestCaseProvider(ABC):
    @abstractmethod
    def download_system_cases(self) -> Iterable[TestCaseData] | bool: ...

    @abstractmethod
    def iter_system_cases(self) -> Iterable[TestCaseFile]: ...

    @property
    def checker(self) -> pathlib.Path | None:
        return None


class Problem(TestCaseProvider):
    def __repr__(self) -> str:
        return f"{self.__class__.__name__}.from_url({self.url!r})"

    def __hash__(self) -> int:  # pragma: no cover
        return hash(self.url) ^ hash(type(self))

    def __eq__(self, value: object) -> bool:
        if type(self) is not type(value):
            return False
        return self.url == cast("Problem", value).url

    @property
    @abstractmethod
    def url(self) -> str: ...

    @classmethod
    @abstractmethod
    def from_url(cls, url: str) -> Optional["Problem"]: ...

    @property
    def hash_id(self):
        return hashlib.md5(self.url.encode(), usedforsecurity=False).hexdigest()

    @property
    def problem_directory(self):
        return config.get_problem_cache_dir() / self.hash_id

    @property
    def test_directory(self):
        return self.problem_directory / "test"



import fnmatch
import hashlib
import os
import pathlib
import traceback
from argparse import ArgumentParser
from collections.abc import Generator
from functools import cached_property
from itertools import chain
from logging import getLogger
from typing import Any, Literal

from pydantic import Field, ValidationError

from competitive_verifier import config, git
from competitive_verifier.arg import IncludeExcludeArguments, VerboseArguments
from competitive_verifier.log import GitHubMessageParams
from competitive_verifier.models import (
    AddtionalSource,
    CommandVerification,
    ConstVerification,
    LocalProblemVerification,
    ProblemVerification,
    ResultStatus,
    Verification,
    VerificationFile,
    VerificationInput,
)
from competitive_verifier.util import resolve_referenced_path

from .problem import problem_from_url
from .verify.languages import LanguageEnvironment
from .verify.list import OjVerifyConfig

logger = getLogger(__name__)


def _get_bundled_dir() -> pathlib.Path:
    return config.get_config_dir() / "bundled"


def _write_bundled(content: bytes, *, path: pathlib.Path) -> pathlib.Path:
    """Write bundled code.

    Returns:
        output file path.
    """
    dest_dir = _get_bundled_dir()
    dest_path = dest_dir / path
    dest_path.parent.mkdir(parents=True, exist_ok=True)
    logger.info("bundle_path=%s", dest_path.as_posix())
    dest_path.write_bytes(content)
    return dest_path


class OjResolver:
    include: list[str]
    exclude: list[str]
    config: OjVerifyConfig
    _match_exclude_cache: dict[pathlib.Path, bool]

    def __init__(
        self,
        *,
        include: list[str],
        exclude: list[str],
        config: OjVerifyConfig,
    ) -> None:
        def _remove_slash(s: str):
            s = os.path.normpath(s)
            while len(s) > 1 and s[-1] == os.sep:
                s = s[:-1]
            return s

        self.include = list(map(_remove_slash, include))
        self.exclude = list(map(_remove_slash, exclude))
        self.config = config
        self._match_exclude_cache = {}

    def _match_exclude2(self, paths: list[pathlib.Path]) -> bool:
        if not paths:
            return False
        path = paths.pop()
        cache = self._match_exclude_cache.get(path, None)
        if cache is not None:
            return cache

        for ex in self.exclude:
            if fnmatch.fnmatch(path.as_posix(), ex):
                self._match_exclude_cache[path] = True
                return True
        result = self._match_exclude2(paths)
        self._match_exclude_cache[path] = result
        return result

    def _match_exclude(self, path: pathlib.Path) -> bool:
        paths = list(path.parents)
        paths.reverse()
        paths.append(path)
        return self._match_exclude2(paths)

    @cached_property
    def _lang_dict(self):
        return self.config.get_dict()

    @staticmethod
    def env_to_verifications(
        env: LanguageEnvironment,
        *,
        attr: dict[str, Any],
        path: pathlib.Path,
        basedir: pathlib.Path,
    ) -> Generator[Verification, None, None]:
        if "IGNORE" in attr:
            yield ConstVerification(status=ResultStatus.SKIPPED)
            return

        error_str = attr.get("ERROR")
        try:
            error = float(error_str) if error_str else None
        except ValueError:
            error = None

        tle_str = attr.get("TLE")
        tle = float(tle_str) if tle_str else None

        mle_str = attr.get("MLE")
        mle = float(mle_str) if mle_str else None

        yield from OjResolver._env_to_local_problem_verification(
            env,
            attr=attr,
            path=path,
            basedir=basedir,
            error=error,
            tle=tle,
            mle=mle,
        )
        yield from OjResolver._env_to_problem_verification(
            env,
            attr=attr,
            path=path,
            basedir=basedir,
            error=error,
            tle=tle,
            mle=mle,
        )
        yield from OjResolver._env_to_standalone_verification(
            env,
            attr=attr,
            path=path,
            basedir=basedir,
        )
        yield from OjResolver._env_to_unittest_verification(attr=attr)

    @staticmethod
    def _env_to_problem_verification(
        env: LanguageEnvironment,
        *,
        attr: dict[str, Any],
        path: pathlib.Path,
        basedir: pathlib.Path,
        error: float | None,
        tle: float | None,
        mle: float | None,
    ) -> Generator[Verification, None, None]:
        url = attr.get("PROBLEM")

        if not url:
            return

        problem = problem_from_url(url)
        if problem is None:
            logger.error(
                'The URL "%s" is not supported',
                url,
                extra={"github": GitHubMessageParams()},
            )
            yield ConstVerification(status=ResultStatus.FAILURE)
        else:
            tempdir = problem.problem_directory
            yield ProblemVerification(
                name=env.name,
                command=env.get_execute_command(path, basedir=basedir, tempdir=tempdir),
                compile=env.get_compile_command(path, basedir=basedir, tempdir=tempdir),
                problem=url,
                error=error,
                tle=tle,
                mle=mle,
            )

    @staticmethod
    def _env_to_local_problem_verification(
        env: LanguageEnvironment,
        *,
        attr: dict[str, Any],
        path: pathlib.Path,
        basedir: pathlib.Path,
        error: float | None,
        tle: float | None,
        mle: float | None,
    ) -> Generator[Verification, None, None]:
        casedir = attr.get("LOCALCASE")
        if casedir is None:
            return
        casedir = resolve_referenced_path(casedir, basedir=path.parent)
        if casedir is None:
            return

        tempdir = (
            config.get_cache_dir()
            / "localcase"
            / hashlib.md5(
                path.as_posix().encode("utf-8"), usedforsecurity=False
            ).hexdigest()
        )
        yield LocalProblemVerification(
            name=env.name,
            command=env.get_execute_command(path, basedir=basedir, tempdir=tempdir),
            compile=env.get_compile_command(path, basedir=basedir, tempdir=tempdir),
            input=casedir,
            error=error,
            tle=tle,
            mle=mle,
            tempdir=tempdir,
        )

    @staticmethod
    def _env_to_standalone_verification(
        env: LanguageEnvironment,
        *,
        attr: dict[str, Any],
        path: pathlib.Path,
        basedir: pathlib.Path,
    ) -> Generator[Verification, None, None]:
        if attr.get("STANDALONE") is None:
            return

        tempdir = (
            config.get_cache_dir()
            / "standalone"
            / hashlib.md5(
                path.as_posix().encode("utf-8"), usedforsecurity=False
            ).hexdigest()
        )
        yield CommandVerification(
            name=env.name,
            command=env.get_execute_command(path, basedir=basedir, tempdir=tempdir),
            compile=env.get_compile_command(path, basedir=basedir, tempdir=tempdir),
            tempdir=tempdir,
        )

    @staticmethod
    def _env_to_unittest_verification(
        *,
        attr: dict[str, Any],
    ) -> Generator[Verification, None, None]:
        unit_test_envvar = attr.get("UNITTEST")
        if not unit_test_envvar:
            return
        var = os.getenv(unit_test_envvar)
        if var is None:
            logger.warning("UNITTEST envvar %s is not defined.", unit_test_envvar)
            yield ConstVerification(status=ResultStatus.FAILURE)
        elif var.lower() == "false" or var == "0":
            logger.info("UNITTEST envvar %s=%s is falsy.", unit_test_envvar, var)
            yield ConstVerification(status=ResultStatus.FAILURE)
        else:
            logger.info("UNITTEST envvar %s=%s is truthy.", unit_test_envvar, var)
            yield ConstVerification(status=ResultStatus.SUCCESS)

    def resolve(self, *, bundle: bool) -> VerificationInput:
        files: dict[pathlib.Path, VerificationFile] = {}
        basedir = pathlib.Path.cwd()

        for path in git.ls_files(*self.include):
            if self._match_exclude(path):
                logger.debug("exclude=%s", path)
                continue

            language = self._lang_dict.get(path.suffix)
            if language is None:
                continue

            deps = set(git.ls_files(*language.list_dependencies(path, basedir=basedir)))
            attr = language.list_attributes(path, basedir=basedir)

            additonal_sources: list[AddtionalSource] = []
            if bundle:
                try:
                    bundled_code = language.bundle(path, basedir=basedir)
                    if bundled_code:
                        dest_path = _write_bundled(bundled_code, path=path)
                        additonal_sources.append(
                            AddtionalSource(name="bundled", path=dest_path)
                        )
                except Exception:  # noqa: BLE001
                    dest_path = _write_bundled(
                        traceback.format_exc().encode(), path=path
                    )
                    additonal_sources.append(
                        AddtionalSource(name="bundle error", path=dest_path)
                    )

            verifications = list(
                chain.from_iterable(
                    self.env_to_verifications(vs, attr=attr, path=path, basedir=basedir)
                    for vs in language.list_environments(path, basedir=basedir)
                )
            )
            files[path] = VerificationFile(
                dependencies=deps,
                verification=verifications,
                document_attributes=attr,
                additonal_sources=additonal_sources,
            )
        return VerificationInput(files=files)


class OjResolve(IncludeExcludeArguments, VerboseArguments):
    subcommand: Literal["oj-resolve"] = Field(
        default="oj-resolve",
        description="Create verify_files json using `oj-verify`",
    )
    bundle: bool = True
    config: pathlib.Path | OjVerifyConfig | None = None

    @classmethod
    def add_parser(cls, parser: ArgumentParser):
        super().add_parser(parser)
        parser.add_argument(
            "--no-bundle",
            dest="bundle",
            action="store_false",
            help="Disable bundle",
        )
        parser.add_argument(
            "--config",
            help="config.toml",
            type=pathlib.Path,
        )

    def to_resolver(self) -> OjResolver:
        if self.config is None:
            logger.info("no config file")
            config = OjVerifyConfig()
        elif not isinstance(self.config, OjVerifyConfig):
            config_path = pathlib.Path(self.config)
            try:
                with config_path.open("rb") as fp:
                    config = OjVerifyConfig.load(fp)
                    logger.info("config file loaded: %s: %s", config_path, config)
            except ValidationError:
                logger.exception(
                    "config file validation error",
                    extra={"github": GitHubMessageParams(file=config_path)},
                )
                config = OjVerifyConfig()
        else:
            config = self.config

        return OjResolver(
            include=self.include,
            exclude=self.exclude,
            config=config,
        )

    def _run(self) -> bool:
        logger.debug("arguments:%s", self)
        resolved = self.to_resolver().resolve(bundle=self.bundle)
        print(resolved.model_dump_json(exclude_none=True))
        return True



import os
import pathlib
import sys
from collections.abc import Sequence
from dataclasses import dataclass
from typing import Any, NamedTuple
from unittest.mock import PropertyMock

import pytest
from pydantic import TypeAdapter
from pytest_mock import MockerFixture

from competitive_verifier.models import (
    CommandVerification,
    ConstVerification,
    LocalProblemVerification,
    Problem,
    ProblemVerification,
    ResultStatus,
    ShellCommand,
    Verification,
)
from competitive_verifier.models.verification import BaseProblemVerification
from competitive_verifier.oj import LocalProblem, problem_from_url
from competitive_verifier.oj.oj_test import OjTestArguments


def from_url_force(url: str):
    p = problem_from_url(url)
    assert p
    return p


@dataclass
class DataVerificationParams:
    default_tle: float | None = 22
    default_mle: float | None = 128


test_command_union_json_params: list[tuple[Verification, str, str]] = [
    (
        ConstVerification(status=ResultStatus.SUCCESS),
        '{"type": "const","status":"success"}',
        '{"type":"const","status":"success"}',
    ),
    (
        ConstVerification(status=ResultStatus.SUCCESS, name="foo"),
        '{"type": "const","name":"foo","status":"success"}',
        '{"name":"foo","type":"const","status":"success"}',
    ),
    (
        ConstVerification(status=ResultStatus.FAILURE),
        '{"type": "const","status":"failure"}',
        '{"type":"const","status":"failure"}',
    ),
    (
        ConstVerification(status=ResultStatus.SKIPPED),
        '{"type": "const","status":"skipped"}',
        '{"type":"const","status":"skipped"}',
    ),
    (
        ConstVerification(status=ResultStatus.SUCCESS),
        '{"type": "const","status":"success"}',
        '{"type":"const","status":"success"}',
    ),
    (
        CommandVerification(command="ls ~"),
        '{"type":"command","command":"ls ~"}',
        '{"type":"command","command":"ls ~"}',
    ),
    (
        CommandVerification(compile="cat LICENSE", command="ls ~"),
        '{"type": "command", "compile": "cat LICENSE", "command": "ls ~"}',
        '{"type":"command","command":"ls ~","compile":"cat LICENSE"}',
    ),
    (
        ProblemVerification(command="ls ~", problem="https://example.com"),
        '{"type":"problem","problem":"https://example.com","command":"ls ~"}',
        '{"command":"ls ~","type":"problem","problem":"https://example.com"}',
    ),
    (
        ProblemVerification(name="bar", command="ls ~", problem="https://example.com"),
        '{"type":"problem","name":"bar","problem":"https://example.com","command":"ls ~"}',
        '{"name":"bar","command":"ls ~","type":"problem","problem":"https://example.com"}',
    ),
    (
        ProblemVerification(
            compile="cat LICENSE", command="ls ~", problem="https://example.com"
        ),
        '{"type": "problem",  "problem":"https://example.com", "compile": "cat LICENSE", "command": "ls ~"}',
        '{"command":"ls ~","compile":"cat LICENSE","type":"problem","problem":"https://example.com"}',
    ),
    (
        ProblemVerification(
            compile="cat LICENSE",
            command="ls ~",
            problem="https://example.com",
            error=1e-6,
            tle=2,
        ),
        '{"type": "problem",  "problem":"https://example.com", "compile": "cat LICENSE", "error": 0.000001, "tle": 2, "command": "ls ~"}',
        '{"command":"ls ~","compile":"cat LICENSE","error":1e-6,"tle":2.0,"type":"problem","problem":"https://example.com"}',
    ),
    (
        LocalProblemVerification(
            compile="cat LICENSE",
            command="ls ~",
            input=pathlib.Path("/root/cases-dir"),
            error=1e-6,
            tle=2,
        ),
        '{"type": "local",  "input":"/root/cases-dir", "compile": "cat LICENSE", "error": 0.000001, "tle": 2, "command": "ls ~"}',
        '{"command":"ls ~","compile":"cat LICENSE","error":1e-6,"tle":2.0,"type":"local","input":"/root/cases-dir"}',
    ),
]


@pytest.mark.parametrize(
    ("obj", "raw_json", "output_json"),
    test_command_union_json_params,
    ids=lambda s: s[:20] if isinstance(s, str) else s,
)
def test_command_union_json(
    obj: Verification,
    raw_json: str,
    output_json: str,
):
    assert obj == type(obj).model_validate_json(raw_json)
    assert obj.model_dump_json(exclude_none=True) == output_json

    assert obj == TypeAdapter(Verification).validate_json(raw_json)


def test_const_verification(mocker: MockerFixture):
    mock_exec_command = mocker.patch("subprocess.run")
    obj = ConstVerification(status=ResultStatus.SUCCESS)

    assert obj.run() == ResultStatus.SUCCESS
    assert obj.run_compile_command()
    mock_exec_command.assert_not_called()


test_run_params: list[tuple[Verification, tuple[str | list[str]], dict[str, Any]]] = [
    (
        CommandVerification(command="ls ~"),
        ("ls ~",),
        {
            "shell": True,
            "text": True,
            "check": False,
            "cwd": None,
            "env": None,
            "capture_output": False,
            "encoding": sys.stdout.encoding,
        },
    ),
    (
        CommandVerification(compile="cat LICENSE", command="ls ~"),
        ("ls ~",),
        {
            "shell": True,
            "text": True,
            "check": False,
            "cwd": None,
            "env": None,
            "capture_output": False,
            "encoding": sys.stdout.encoding,
        },
    ),
    (
        CommandVerification(command=["ls", "~"]),
        (["ls", "~"],),
        {
            "shell": False,
            "text": True,
            "check": False,
            "cwd": None,
            "env": None,
            "capture_output": False,
            "encoding": sys.stdout.encoding,
        },
    ),
    (
        CommandVerification(command=["ls", "~"]),
        (["ls", "~"],),
        {
            "shell": False,
            "text": True,
            "check": False,
            "cwd": None,
            "env": None,
            "capture_output": False,
            "encoding": sys.stdout.encoding,
        },
    ),
    (
        CommandVerification(
            command=ShellCommand(command="ls ~", cwd=pathlib.Path("/foo")),
        ),
        ("ls ~",),
        {
            "shell": True,
            "text": True,
            "check": False,
            "cwd": pathlib.Path("/foo"),
            "env": None,
            "capture_output": False,
            "encoding": sys.stdout.encoding,
        },
    ),
    (
        CommandVerification(
            command=ShellCommand(command=["ls", "~"], cwd=pathlib.Path("~")),
        ),
        (["ls", "~"],),
        {
            "shell": False,
            "text": True,
            "check": False,
            "cwd": pathlib.Path("~"),
            "env": None,
            "capture_output": False,
            "encoding": sys.stdout.encoding,
        },
    ),
]


@pytest.mark.parametrize(
    ("obj", "args", "kwargs"),
    test_run_params,
)
def test_run(
    obj: Verification,
    args: Sequence[Any],
    kwargs: dict[str, Any],
    mocker: MockerFixture,
):
    mock_exec_command = mocker.patch("subprocess.run")
    obj.run(DataVerificationParams())
    mock_exec_command.assert_called_once_with(*args, **kwargs)


test_run_with_env_params: list[
    tuple[Verification, tuple[str | list[str]], dict[str, Any], dict[str, str]]
] = [
    (
        CommandVerification(
            command=ShellCommand(command="ls ~", env={"TOKEN": "DUMMY"}),
        ),
        ("ls ~",),
        {
            "shell": True,
            "text": True,
            "check": False,
            "cwd": None,
            "capture_output": False,
            "encoding": sys.stdout.encoding,
        },
        {"TOKEN": "DUMMY"},
    ),
    (
        CommandVerification(
            command=ShellCommand(command=["ls", "~"], env={"TOKEN": "DUMMY"}),
        ),
        (["ls", "~"],),
        {
            "shell": False,
            "text": True,
            "check": False,
            "cwd": None,
            "capture_output": False,
            "encoding": sys.stdout.encoding,
        },
        {"TOKEN": "DUMMY"},
    ),
]


@pytest.mark.parametrize(
    ("obj", "args", "kwargs", "env"),
    test_run_with_env_params,
)
def test_run_with_env(
    obj: Verification,
    args: Sequence[Any],
    kwargs: dict[str, Any],
    env: dict[str, str],
    mocker: MockerFixture,
):
    mock_exec_command = mocker.patch("subprocess.run")
    obj.run(DataVerificationParams())
    mock_exec_command.assert_called_once()
    assert mock_exec_command.call_args[0] == args

    call_kwargs: dict[str, Any] = mock_exec_command.call_args[1]
    assert dict(call_kwargs["env"].items() & env.items()) == env
    assert len(call_kwargs["env"]) > len(env)
    del call_kwargs["env"]
    assert call_kwargs == kwargs


test_run_problem_command_params: list[
    tuple[BaseProblemVerification, OjTestArguments]
] = [
    (
        ProblemVerification(
            command="ls ~",
            problem="http://judge.u-aizu.ac.jp/onlinejudge/description.jsp?id=1169",
        ),
        OjTestArguments(
            problem=from_url_force(
                "http://judge.u-aizu.ac.jp/onlinejudge/description.jsp?id=1169"
            ),
            command="ls ~",
            tle=22.0,
            error=None,
            mle=128,
            env=None,
        ),
    ),
    (
        ProblemVerification(
            compile="cat LICENSE",
            command="ls ~",
            problem="http://judge.u-aizu.ac.jp/onlinejudge/description.jsp?id=1169",
        ),
        OjTestArguments(
            problem=from_url_force(
                "http://judge.u-aizu.ac.jp/onlinejudge/description.jsp?id=1169"
            ),
            command="ls ~",
            tle=22.0,
            error=None,
            mle=128,
            env=None,
        ),
    ),
    (
        ProblemVerification(
            compile="cat LICENSE",
            command="ls ~",
            problem="http://judge.u-aizu.ac.jp/onlinejudge/description.jsp?id=1169",
            error=1e-6,
            tle=2,
            mle=1.2,
        ),
        OjTestArguments(
            problem=from_url_force(
                "http://judge.u-aizu.ac.jp/onlinejudge/description.jsp?id=1169"
            ),
            command="ls ~",
            tle=2.0,
            error=1e-06,
            mle=1.2,
            env=None,
        ),
    ),
    (
        ProblemVerification(
            compile=ShellCommand(command="cat LICENSE", env={"ARG": "VAR"}),
            command="ls ~",
            problem="https://judge.yosupo.jp/problem/aplusb",
            error=1e-6,
            tle=2,
        ),
        OjTestArguments(
            problem=from_url_force("https://judge.yosupo.jp/problem/aplusb"),
            command="ls ~",
            tle=2.0,
            error=1e-06,
            mle=128,
            env=None,
        ),
    ),
    (
        ProblemVerification(
            compile="cat LICENSE",
            command=ShellCommand(command="ls ~", env={"ARG": "VAR"}),
            problem="https://judge.yosupo.jp/problem/aplusb",
            error=1e-6,
            tle=2,
        ),
        OjTestArguments(
            problem=from_url_force("https://judge.yosupo.jp/problem/aplusb"),
            command="ls ~",
            tle=2.0,
            error=1e-06,
            mle=128,
            env={"ARG": "VAR"},
        ),
    ),
    (
        LocalProblemVerification(
            compile="cat LICENSE",
            command=ShellCommand(command="ls ~", env={"ARG": "VAR"}),
            input=pathlib.Path("/root/testcases"),
            error=1e-6,
            tle=2,
        ),
        OjTestArguments(
            problem=LocalProblem(path=pathlib.Path("/root/testcases")),
            command="ls ~",
            tle=2.0,
            error=1e-06,
            mle=128,
            env={"ARG": "VAR"},
        ),
    ),
]


@pytest.mark.parametrize(
    ("obj", "args"),
    test_run_problem_command_params,
)
def test_run_problem_command(
    obj: ProblemVerification,
    args: OjTestArguments,
    mocker: MockerFixture,
):
    mocker.patch(
        "competitive_verifier.oj.problem.LibraryCheckerProblem.checker_exe_name",
        "mockcheck",
    )
    patch = mocker.patch("competitive_verifier.oj.oj_test._run")

    mocker.patch.object(
        Problem,
        "problem_directory",
        new_callable=PropertyMock,
        return_value=pathlib.Path("/any/"),
    )

    obj.run(DataVerificationParams())
    patch.assert_called_once_with(args)


test_run_compile_params: list[
    tuple[Verification, str | list[str] | None, dict[str, Any] | None]
] = [
    (
        CommandVerification(command="ls ~"),
        None,
        None,
    ),
    (
        CommandVerification(compile="cat LICENSE", command="ls ~"),
        "cat LICENSE",
        {
            "shell": True,
            "text": True,
            "check": False,
            "env": None,
            "cwd": None,
            "capture_output": False,
            "encoding": sys.stdout.encoding,
        },
    ),
    (
        ProblemVerification(command="ls ~", problem="https://example.com"),
        None,
        None,
    ),
    (
        ProblemVerification(
            compile="cat LICENSE", command="ls ~", problem="https://example.com"
        ),
        "cat LICENSE",
        {
            "shell": True,
            "text": True,
            "check": False,
            "env": None,
            "cwd": None,
            "capture_output": False,
            "encoding": sys.stdout.encoding,
        },
    ),
    (
        ProblemVerification(
            compile="cat LICENSE",
            command="ls ~",
            problem="https://example.com",
            error=1e-6,
            tle=2,
        ),
        "cat LICENSE",
        {
            "shell": True,
            "text": True,
            "check": False,
            "env": None,
            "cwd": None,
            "capture_output": False,
            "encoding": sys.stdout.encoding,
        },
    ),
    (
        ProblemVerification(
            compile=["cat", "LICENSE"],
            command="ls ~",
            problem="https://example.com",
            error=1e-6,
            tle=2,
        ),
        ["cat", "LICENSE"],
        {
            "shell": False,
            "text": True,
            "check": False,
            "env": None,
            "cwd": None,
            "capture_output": False,
            "encoding": sys.stdout.encoding,
        },
    ),
    (
        ProblemVerification(
            compile=ShellCommand(
                command=["cat", "LICENSE"],
                cwd=pathlib.Path("~/foo"),
            ),
            command="ls ~",
            problem="https://example.com",
            error=1e-6,
            tle=2,
        ),
        ["cat", "LICENSE"],
        {
            "shell": False,
            "text": True,
            "check": False,
            "env": None,
            "cwd": pathlib.Path("~/foo"),
            "capture_output": False,
            "encoding": sys.stdout.encoding,
        },
    ),
    (
        LocalProblemVerification(
            compile=ShellCommand(
                command=["cat", "LICENSE"],
                cwd=pathlib.Path("~/foo"),
            ),
            command="ls ~",
            input=pathlib.Path("/root/testcases"),
            error=1e-6,
            tle=2,
        ),
        ["cat", "LICENSE"],
        {
            "shell": False,
            "text": True,
            "check": False,
            "env": None,
            "cwd": pathlib.Path("~/foo"),
            "capture_output": False,
            "encoding": sys.stdout.encoding,
        },
    ),
]


@pytest.mark.parametrize(
    ("obj", "args", "kwargs"),
    test_run_compile_params,
)
def test_run_compile(
    obj: Verification,
    args: Sequence[Any] | None,
    kwargs: dict[str, Any] | None,
    mocker: MockerFixture,
):
    mock_exec_command = mocker.patch("subprocess.run")
    obj.run_compile_command(DataVerificationParams())
    if args is None:
        mock_exec_command.assert_not_called()
    else:
        assert kwargs
        mock_exec_command.assert_called_once_with(args, **kwargs)


test_params_run_params: list[tuple[Verification, str | None]] = [
    (ConstVerification(status=ResultStatus.SUCCESS), None),
    (CommandVerification(command="true"), None),
    (
        ProblemVerification(command="true", problem="http://example.com"),
        "ProblemVerification.run requires VerificationParams",
    ),
]


@pytest.mark.parametrize(
    ("obj", "error_message"),
    test_params_run_params,
)
def test_params_run(
    obj: Verification,
    error_message: str | None,
    mocker: MockerFixture,
):
    mocker.patch("subprocess.run")
    if error_message:
        with pytest.raises(ValueError, match=error_message) as e:
            obj.run()
        assert e.value.args == (error_message,)
    else:
        obj.run()


@pytest.mark.parametrize(
    "status", [ResultStatus.FAILURE, ResultStatus.SKIPPED, ResultStatus.SUCCESS]
)
def test_run_const_verification(status: ResultStatus):
    obj = ConstVerification(name="name", status=status)
    assert obj.run_compile_command()
    assert obj.run() == status


class Return(NamedTuple):
    returncode: int


class TestCommandVerification:
    def test_command(self, mocker: MockerFixture):
        mockrun = mocker.patch("subprocess.run", return_value=Return(0))
        obj = CommandVerification(name="name", command="echo 1")

        assert obj.run_compile_command()
        assert obj.run() == ResultStatus.SUCCESS
        mockrun.assert_called_once_with(
            "echo 1",
            capture_output=False,
            check=False,
            cwd=None,
            encoding="utf-8",
            env=None,
            shell=True,
            text=True,
        )

    def test_command_failure(self, mocker: MockerFixture):
        mockrun = mocker.patch("subprocess.run", return_value=Return(1))
        obj = CommandVerification(name="name", command=["echo", "1"])

        assert obj.run_compile_command()
        assert obj.run() == ResultStatus.FAILURE
        mockrun.assert_called_once_with(
            ["echo", "1"],
            capture_output=False,
            check=False,
            cwd=None,
            encoding="utf-8",
            env=None,
            shell=False,
            text=True,
        )

    def test_command_and_compile(self, mocker: MockerFixture):
        mocker.patch.dict(os.environ, {"DEFAULT": "dd"}, clear=True)
        mockrun = mocker.patch("subprocess.run", return_value=Return(0))
        obj = CommandVerification(
            name="name",
            compile="echo 1",
            command=ShellCommand(
                command="ls .",
                env={"MOCK": "mocker"},
                cwd=pathlib.Path("/root"),
            ),
        )

        assert obj.run_compile_command()
        mockrun.assert_called_once_with(
            "echo 1",
            capture_output=False,
            check=False,
            cwd=None,
            encoding="utf-8",
            env=None,
            shell=True,
            text=True,
        )
        mockrun.reset_mock()

        assert obj.run() == ResultStatus.SUCCESS
        mockrun.assert_called_once_with(
            "ls .",
            capture_output=False,
            check=False,
            cwd=pathlib.Path("/root"),
            encoding="utf-8",
            env={"DEFAULT": "dd", "MOCK": "mocker"},
            shell=True,
            text=True,
        )

    def test_tempdir(self, mocker: MockerFixture):
        mocker.patch.dict(os.environ, {"DEFAULT": "dd"}, clear=True)
        mockrun = mocker.patch("subprocess.run", return_value=Return(0))

        mockmkdir = mocker.patch.object(pathlib.Path, "mkdir")

        obj = CommandVerification(
            name="name",
            compile="echo 1",
            command=ShellCommand(
                command="ls .",
                env={"MOCK": "mocker"},
                cwd=pathlib.Path("/root"),
            ),
            tempdir=pathlib.Path("/root/tmp"),
        )

        assert obj.run_compile_command()
        mockmkdir.assert_called_once_with(exist_ok=True, parents=True)
        mockrun.assert_called_once_with(
            "echo 1",
            capture_output=False,
            check=False,
            cwd=None,
            encoding="utf-8",
            env=None,
            shell=True,
            text=True,
        )
        mockmkdir.reset_mock()
        mockrun.reset_mock()

        assert obj.run() == ResultStatus.SUCCESS
        mockmkdir.assert_called_once_with(exist_ok=True, parents=True)
        mockrun.assert_called_once_with(
            "ls .",
            capture_output=False,
            check=False,
            cwd=pathlib.Path("/root"),
            encoding="utf-8",
            env={"DEFAULT": "dd", "MOCK": "mocker"},
            shell=True,
            text=True,
        )


def test_problem_verification_invalid_url():
    assert (
        ProblemVerification(
            problem="https://example.com/notfound", command="false"
        ).run(DataVerificationParams())
        == ResultStatus.FAILURE
    )


@pytest.mark.allow_mkdir
def test_local_problem_verification_tempdir(testtemp: pathlib.Path):
    tempdir = testtemp / "foo/bar/baz"
    assert not tempdir.is_dir()

    LocalProblemVerification(
        command="ls ~",
        input=pathlib.Path("/root/cases-dir"),
        tempdir=tempdir,
    ).run_compile_command()

    assert tempdir.is_dir()



import os
import pathlib
import shutil
import subprocess

import pytest
from pytest_mock import MockerFixture

from competitive_verifier import app
from competitive_verifier.models import (
    ProblemVerification,
    VerificationFile,
    VerificationInput,
)
from competitive_verifier.models import TestCaseFile as Case
from competitive_verifier.oj.problem import LibraryCheckerProblem, problem_from_url

from .types import FilePaths


@pytest.fixture(scope="session")
def clean_download_dir(file_paths: FilePaths) -> pathlib.Path:
    d = file_paths.dest_root / "download"

    if d.exists():  # pragma: no cover
        shutil.rmtree(d)

    return d


@pytest.fixture
def download_config_dir(
    mocker: MockerFixture,
    request: pytest.FixtureRequest,
    clean_download_dir: pathlib.Path,
) -> pathlib.Path:
    func_name: str = request.function.__name__
    assert func_name.startswith("test_")
    d = clean_download_dir / func_name[len("test_") :]

    mocker.patch.dict(
        os.environ,
        {"COMPETITIVE_VERIFY_CONFIG_PATH": d.as_posix()},
    )
    return d


@pytest.mark.usefixtures("additional_path")
@pytest.mark.order(-1)
class TestCommandDownload:
    @pytest.mark.integration
    @pytest.mark.skipif(
        not bool(shutil.which("g++") or os.getenv("CXX")),
        reason="g++ is not found",
    )
    def test_library_checker(
        self,
        download_config_dir: pathlib.Path,
        testtemp: pathlib.Path,
    ):
        url = "https://judge.yosupo.jp/problem/aplusb"

        (testtemp / "verify.json").write_text(
            VerificationInput(
                files={
                    pathlib.Path("lib.txt"): VerificationFile(
                        dependencies={pathlib.Path("test.txt")}
                    ),
                    pathlib.Path("test.txt"): VerificationFile(
                        verification=ProblemVerification(
                            name="aplusb", command="true", problem=url
                        )
                    ),
                }
            ).model_dump_json()
        )

        problem = problem_from_url(url)
        assert problem is not None
        assert isinstance(problem, LibraryCheckerProblem)

        problem_path = problem.problem_directory
        assert download_config_dir.absolute().is_relative_to(pathlib.Path.cwd())
        assert problem_path.is_relative_to(download_config_dir)

        parsed = app.ArgumentParser().parse(
            ["download", "--verify-json", str(testtemp / "verify.json")]
        )
        assert isinstance(parsed, app.Download)
        assert parsed.run()
        assert problem_path.exists()

        testcases = list(problem.iter_system_cases())
        assert testcases == [
            Case(
                name=name,
                input_path=problem.source_directory / f"in/{name}.in",
                output_path=problem.source_directory / f"out/{name}.out",
            )
            for name in [
                "example_00",
                "example_01",
                "random_00",
                "random_01",
                "random_02",
                "random_03",
                "random_04",
                "random_05",
                "random_06",
                "random_07",
                "random_08",
                "random_09",
            ]
        ]

        expected_outputs = {
            "example_00": b"6912\n",
            "example_01": b"2000000000\n",
            "random_00": b"348927966\n",
            "random_01": b"385703343\n",
            "random_02": b"1133946337\n",
            "random_03": b"1132392878\n",
            "random_04": b"1144896831\n",
            "random_05": b"393933893\n",
            "random_06": b"899846272\n",
            "random_07": b"507618092\n",
            "random_08": b"340723397\n",
            "random_09": b"1198300249\n",
        }

        assert {
            t.name: t.output_path.read_bytes() for t in testcases
        } == expected_outputs

        for t in testcases:
            checker = problem.checker
            assert (
                subprocess.run(
                    [
                        str(checker),
                        str(t.input_path),
                        str(t.output_path),
                        str(t.output_path),
                    ],
                    check=False,
                ).returncode
                == 0
            )

    @pytest.mark.integration
    @pytest.mark.skipif(
        not bool(os.getenv("YUKICODER_TOKEN")),
        reason="$YUKICODER_TOKEN is not found",
    )
    @pytest.mark.parametrize(
        "url",
        ["https://yukicoder.me/problems/4573", "https://yukicoder.me/problems/no/1088"],
    )
    def test_yukicoder(self, url: str, download_config_dir: pathlib.Path):
        problem = problem_from_url(url)
        assert problem is not None

        problem_path = problem.problem_directory
        assert download_config_dir.absolute().is_relative_to(pathlib.Path.cwd())
        assert problem_path.is_relative_to(download_config_dir)

        parsed = app.ArgumentParser().parse(["download", url])
        assert isinstance(parsed, app.Download)
        assert parsed.run()

        testcases = list(problem.iter_system_cases())
        assert testcases == [
            Case(
                name=name,
                input_path=problem_path / f"test/{name}.in",
                output_path=problem_path / f"test/{name}.out",
            )
            for name in [
                "01_sample_01",
                "01_sample_02",
                "input1",
                "input10",
                "input2",
                "input3",
                "input4",
                "input5",
                "input6",
                "input7",
                "input8",
                "input9",
            ]
        ]

        expected_outputs = {
            "input1": b"103\n",
            "input2": b"112\n",
            "input3": b"74\n",
            "input4": b"45\n",
            "input5": b"96\n",
            "input6": b"129\n",
            "input7": b"65\n",
            "input8": b"92\n",
            "input9": b"65\n",
            "input10": b"130\n",
            "01_sample_01": b"46\n",
            "01_sample_02": b"72\n",
        }
        assert {
            t.name: t.output_path.read_bytes() for t in testcases
        } == expected_outputs

    @pytest.mark.integration
    def test_aizu_onlinejudge(self, download_config_dir: pathlib.Path):
        urls = [
            "https://onlinejudge.u-aizu.ac.jp/courses/lesson/2/ITP1/1/ITP1_1_A",
            "https://onlinejudge.u-aizu.ac.jp/problems/ITP1_1_A",
            "https://judge.u-aizu.ac.jp/onlinejudge/description.jsp?id=ITP1_1_A&lang=jp",
        ]
        parsed = app.ArgumentParser().parse(["download", *urls])
        assert isinstance(parsed, app.Download)
        assert parsed.run()

        for url in urls:
            problem = problem_from_url(url)
            assert problem is not None

            problem_path = problem.problem_directory
            assert download_config_dir.absolute().is_relative_to(pathlib.Path.cwd())
            assert problem_path.is_relative_to(download_config_dir)

            testcases = list(problem.iter_system_cases())
            assert testcases == [
                Case(
                    name="judge_data",
                    input_path=problem_path / "test/judge_data.in",
                    output_path=problem_path / "test/judge_data.out",
                )
            ]

            expected_outputs = {"judge_data": b"Hello World\n"}
            assert {
                t.name: t.output_path.read_bytes() for t in testcases
            } == expected_outputs

    @pytest.mark.integration
    def test_aizu_onlinejudge_arena(self, download_config_dir: pathlib.Path):
        url = "https://onlinejudge.u-aizu.ac.jp/services/room.html#RitsCamp19Day2/problems/A"

        problem = problem_from_url(url)
        assert problem is not None

        problem_path = problem.problem_directory
        assert download_config_dir.absolute().is_relative_to(pathlib.Path.cwd())
        assert problem_path.is_relative_to(download_config_dir)

        parsed = app.ArgumentParser().parse(["download", url])
        assert isinstance(parsed, app.Download)
        assert parsed.run()

        testcases = list(problem.iter_system_cases())
        assert testcases == [
            Case(
                name=name,
                input_path=problem_path / f"test/{name}.in",
                output_path=problem_path / f"test/{name}.out",
            )
            for name in [
                "00_sample_01",
                "00_sample_02",
                "01_random_01",
                "01_random_02",
                "01_random_03",
                "01_random_04",
                "01_random_05",
                "01_random_06",
                "01_random_07",
                "01_random_08",
                "01_random_09",
                "01_random_10",
                "01_random_11",
                "01_random_12",
                "01_random_13",
                "01_random_14",
                "01_random_15",
                "01_random_16",
                "01_random_17",
                "01_random_18",
                "01_random_19",
                "01_random_20",
            ]
        ]

        expected_outputs = {
            "00_sample_01": b"A\n",
            "00_sample_02": b"C\n",
            "01_random_01": b"B\n",
            "01_random_02": b"A\n",
            "01_random_03": b"C\n",
            "01_random_04": b"C\n",
            "01_random_05": b"B\n",
            "01_random_06": b"A\n",
            "01_random_07": b"C\n",
            "01_random_08": b"C\n",
            "01_random_09": b"A\n",
            "01_random_10": b"B\n",
            "01_random_11": b"C\n",
            "01_random_12": b"B\n",
            "01_random_13": b"A\n",
            "01_random_14": b"C\n",
            "01_random_15": b"A\n",
            "01_random_16": b"A\n",
            "01_random_17": b"C\n",
            "01_random_18": b"A\n",
            "01_random_19": b"C\n",
            "01_random_20": b"C\n",
        }
        assert {
            t.name: t.output_path.read_bytes() for t in testcases
        } == expected_outputs



import pytest

from competitive_verifier.oj.problem import (
    _normpath,  # pyright: ignore[reportPrivateUsage]
    problem_from_url,
)

test_normpath_params: list[tuple[str, str]] = [
    ("hoge/foo/bar", "hoge/foo/bar"),
    ("/foo/bar", "/foo/bar"),
    ("//foo/bar", "/foo/bar"),
]


@pytest.mark.parametrize(
    ("path", "expected"),
    test_normpath_params,
    ids=[t[0] for t in test_normpath_params],
)
def test_normpath(path: str, expected: str):
    assert _normpath(path) == expected


test_problem_repr_params = [
    (
        "https://onlinejudge.u-aizu.ac.jp/courses/lesson/2/ITP1/1/ITP1_1_A",
        "AOJProblem.from_url('http://judge.u-aizu.ac.jp/onlinejudge/description.jsp?id=ITP1_1_A')",
    ),
    (
        "https://onlinejudge.u-aizu.ac.jp/problems/ITP1_1_A",
        "AOJProblem.from_url('http://judge.u-aizu.ac.jp/onlinejudge/description.jsp?id=ITP1_1_A')",
    ),
    (
        "https://judge.u-aizu.ac.jp/onlinejudge/description.jsp?id=ITP1_1_A&lang=jp",
        "AOJProblem.from_url('http://judge.u-aizu.ac.jp/onlinejudge/description.jsp?id=ITP1_1_A')",
    ),
    (
        "https://onlinejudge.u-aizu.ac.jp/services/room.html#RitsCamp19Day2/problems/A",
        "AOJArenaProblem.from_url('https://onlinejudge.u-aizu.ac.jp/services/room.html#RitsCamp19Day2/problems/A')",
    ),
    (
        "https://old.yosupo.jp/problem/aplusb",
        "LibraryCheckerProblem.from_url('https://judge.yosupo.jp/problem/aplusb')",
    ),
    (
        "https://judge.yosupo.jp/problem/aplusb",
        "LibraryCheckerProblem.from_url('https://judge.yosupo.jp/problem/aplusb')",
    ),
    (
        "http://old.yosupo.jp/problem/aplusb",
        "LibraryCheckerProblem.from_url('https://judge.yosupo.jp/problem/aplusb')",
    ),
    (
        "http://judge.yosupo.jp/problem/aplusb",
        "LibraryCheckerProblem.from_url('https://judge.yosupo.jp/problem/aplusb')",
    ),
    (
        "https://yukicoder.me/problems/4573",
        "YukicoderProblem.from_url('https://yukicoder.me/problems/4573')",
    ),
    (
        "https://yukicoder.me/problems/no/1088",
        "YukicoderProblem.from_url('https://yukicoder.me/problems/no/1088')",
    ),
    (
        "http://yukicoder.me/problems/4573",
        "YukicoderProblem.from_url('https://yukicoder.me/problems/4573')",
    ),
    (
        "http://yukicoder.me/problems/no/1088",
        "YukicoderProblem.from_url('https://yukicoder.me/problems/no/1088')",
    ),
    (
        "http://yukicoder.me/4573",
        "None",
    ),
]


@pytest.mark.parametrize(
    ("url", "expected"),
    test_problem_repr_params,
    ids=[t[0] for t in test_problem_repr_params],
)
def test_problem_repr(url: str, expected: str):
    assert repr(problem_from_url(url)) == expected



from collections.abc import Mapping
from dataclasses import astuple, dataclass, fields, replace
from logging import LogRecord

from competitive_verifier.log import GitHubMessageParams


@dataclass
class LogComparer:  # noqa: PLW1641
    message: str
    level: int | None = None
    args: tuple[object, ...] | Mapping[str, object] | None = None
    name: str | None = None
    github: GitHubMessageParams | None = None

    def fill(self, other: "LogComparer") -> "LogComparer":
        other_dict = {
            f.name: getattr(other, f.name)
            for f in fields(self)
            if f.name != "github" and getattr(self, f.name) is None
        }
        return replace(self, **other_dict)

    @staticmethod
    def from_record(value: LogRecord) -> "LogComparer":
        return LogComparer(
            message=value.getMessage(),
            level=value.levelno,
            name=value.name,
            args=value.args,
            github=getattr(value, "github", None),
        )

    def __eq__(self, value: object) -> bool:
        if isinstance(value, LogRecord):
            value = self.from_record(value)
            return astuple(self.fill(value)) == astuple(value)
        if type(value) is LogComparer:
            return astuple(self) == astuple(value)
        return NotImplemented



import pathlib

import pytest
from pytest_mock import MockerFixture

from competitive_verifier.documents.main import Docs, get_default_docs_dir
from competitive_verifier.models import VerificationInput, VerifyCommandResult


@pytest.mark.allow_mkdir
def test_get_default_docs_dir(testtemp: pathlib.Path):
    assert get_default_docs_dir() == pathlib.Path(".competitive-verifier/docs")
    (testtemp / ".verify-helper/docs").mkdir(parents=True)
    assert get_default_docs_dir() == pathlib.Path(".verify-helper/docs")
    (testtemp / ".competitive-verifier/docs").mkdir(parents=True)
    assert get_default_docs_dir() == pathlib.Path(".competitive-verifier/docs")


@pytest.mark.parametrize("write_summary", [True, False])
def test_summary(
    write_summary: bool,
    mocker: MockerFixture,
    caplog: pytest.LogCaptureFixture,
    testtemp: pathlib.Path,
):
    caplog.set_level(0)

    dummy_return = object()
    mocker.patch(
        "competitive_verifier.documents.builder.DocumentBuilder.build",
        return_value=dummy_return,
    )
    mock_summary_result = mocker.patch(
        "competitive_verifier.arg.WriteSummaryArguments.write_result"
    )

    (verify_files_json := testtemp / "verify.json").write_text(
        VerificationInput().model_dump_json()
    )

    (testtemp / "result.json").write_text(
        VerifyCommandResult(total_seconds=1).model_dump_json()
    )

    assert (
        Docs(
            verify_files_json=verify_files_json,
            result_json=[testtemp / "result.json"],
            destination=testtemp / "out",
            write_summary=write_summary,
        ).run()
        is dummy_return
    )

    if write_summary:
        mock_summary_result.assert_called_once()
    else:
        mock_summary_result.assert_not_called()



import pathlib
import subprocess
import tempfile
from datetime import datetime, timedelta, timezone

import pytest

from competitive_verifier import git


@pytest.fixture(scope="session")
def mock_repo_make(integration_test_data_dir: pathlib.Path):
    bundle = integration_test_data_dir / "test-repository.bundle"

    with tempfile.TemporaryDirectory() as tmpdir:
        subprocess.run(["git", "clone", str(bundle), tmpdir], check=True)  # noqa: S607
        yield pathlib.Path(tmpdir).resolve()


@pytest.fixture
def mock_repo(mock_repo_make: pathlib.Path, monkeypatch: pytest.MonkeyPatch):
    monkeypatch.chdir(mock_repo_make)
    return mock_repo_make


@pytest.mark.usefixtures("mock_repo")
@pytest.mark.parametrize(
    ("files", "expected"),
    [
        (
            [],
            datetime(2025, 9, 20, 10, 34, 55, tzinfo=timezone(timedelta(hours=-7))),
        ),
        (
            [pathlib.Path("files1")],
            datetime(2025, 9, 20, 3, 21, 21, tzinfo=timezone(timedelta(hours=+9))),
        ),
        (
            [pathlib.Path("files2")],
            datetime(2025, 9, 20, 11, 34, 55, tzinfo=timezone.utc),
        ),
        (
            [pathlib.Path("files1/foo.txt"), pathlib.Path("r1.txt")],
            datetime(2025, 9, 20, 10, 34, 55, tzinfo=timezone(timedelta(hours=-7))),
        ),
        (
            [pathlib.Path("notmatch")],
            datetime.min.replace(tzinfo=timezone.utc),
        ),
    ],
)
def test_get_commit_time(files: list[pathlib.Path], expected: datetime):
    assert git.get_commit_time(files) == expected


@pytest.mark.usefixtures("mock_repo")
@pytest.mark.parametrize(
    ("files", "expected"),
    [
        (
            [],
            {
                pathlib.Path("files1/foo.txt"),
                pathlib.Path("files1/bar.txt"),
                pathlib.Path("files1/baz.txt"),
                pathlib.Path("files2/a.c"),
                pathlib.Path("files2/b.c"),
                pathlib.Path("r1.txt"),
                pathlib.Path("r2.txt"),
            },
        ),
        (
            ["files1"],
            {
                pathlib.Path("files1/foo.txt"),
                pathlib.Path("files1/bar.txt"),
                pathlib.Path("files1/baz.txt"),
            },
        ),
        (
            ["files1/foo.txt", "r1.txt"],
            {
                pathlib.Path("files1/foo.txt"),
                pathlib.Path("r1.txt"),
            },
        ),
    ],
)
def test_ls_files(
    files: list[str],
    expected: set[pathlib.Path],
):
    assert git.ls_files(*files) == expected


def test_get_root_directory(mock_repo: pathlib.Path, monkeypatch: pytest.MonkeyPatch):
    assert git.get_root_directory() == mock_repo
    monkeypatch.chdir("./files1")
    assert git.get_root_directory() == mock_repo



import os
from logging import LogRecord
from typing import TypeGuard, TypeVar

import pytest
from pytest_mock import MockerFixture

from tests import LogComparer


def pytest_addoption(parser: pytest.Parser):
    parser.addoption(
        "--use-prev-dest",
        action="store_true",
        help="Skip deletion of dst_dir.",
    )


def pytest_assertrepr_compare(
    config: pytest.Config,
    op: str,
    left: object,
    right: object,
):
    if op == "==" and _typed_list(left, LogRecord) and _typed_list(right, LogComparer):
        return _asserrepr_log_comparer(config, left, right)
    return None


def _asserrepr_log_comparer(
    config: pytest.Config, left: list[LogRecord], right: list[LogComparer]
) -> list[str] | None:
    size = min(len(left), len(right))
    if size == 0:
        return None
    for i in range(size):
        ll = LogComparer.from_record(left[i])
        left[i] = ll  # pyright: ignore[reportArgumentType, reportCallIssue]
        right[i] = right[i].fill(ll)
    rt = config.hook.pytest_assertrepr_compare(
        config=config, op="==", left=left, right=right
    )
    return rt[0]


TVal = TypeVar("TVal")


def _typed_list(value: object, cls: type[TVal]) -> TypeGuard[list[TVal]]:
    return (
        isinstance(value, list) and all(isinstance(v, cls) for v in value)  # pyright: ignore[reportUnknownVariableType]
    )


@pytest.fixture(autouse=True)
def force_not_github_actions(mocker: MockerFixture):
    mocker.patch.dict(os.environ, {"GITHUB_ACTIONS": ""})



import pathlib
from subprocess import CompletedProcess
from typing import Annotated, Literal, overload

from pydantic import BaseModel, Field
from typing_extensions import TypeAliasType

from competitive_verifier.exec import exec_command


class ShellCommand(BaseModel):
    command: list[str] | str = Field(
        description="Shell command",
    )
    """Shell command
    """

    env: dict[str, str] | None = Field(
        default=None,
        description="Envitonment variables for command",
    )
    """Envitonment variables for command
    """

    cwd: pathlib.Path | None = Field(
        default=None,
        description="The working directory of child process.",
    )
    """The working directory of child process.
    """

    @overload
    def exec_command(
        self,
        *,
        text: Literal[False] = False,
        check: bool = False,
        capture_output: bool = False,
        group_log: bool = False,
    ) -> CompletedProcess[bytes]: ...

    @overload
    def exec_command(
        self,
        *,
        text: Literal[True],
        check: bool = False,
        capture_output: bool = False,
        group_log: bool = False,
    ) -> CompletedProcess[str]: ...

    def exec_command(
        self,
        *,
        text: bool = False,
        check: bool = False,
        capture_output: bool = False,
        group_log: bool = False,
    ) -> CompletedProcess[str] | CompletedProcess[bytes]:
        return exec_command(
            command=self.command,
            env=self.env,
            text=text,
            check=check,
            capture_output=capture_output,
            cwd=self.cwd,
            group_log=group_log,
        )

    @classmethod
    def parse_command_like(cls, cmd: "ShellCommandLike") -> "ShellCommand":
        if isinstance(cmd, (str, list)):
            return ShellCommand(command=cmd)
        return cmd


ShellCommandLike = TypeAliasType(
    "ShellCommandLike",
    Annotated[
        ShellCommand | list[str] | str,
        Field(
            examples=[
                "command",
                ["command", "arg1", "arg2"],
                ShellCommand(
                    command=["command", "arg1", "arg2"],
                    env={"ENVVAR": "DUMMY"},
                    cwd=pathlib.Path("/work"),
                ),
                ShellCommand(
                    command="command",
                    env={"ENVVAR": "DUMMY"},
                    cwd=pathlib.Path("/work"),
                ),
            ]
        ),
    ],
)



import os
import subprocess
import sys
from contextlib import nullcontext
from logging import getLogger
from typing import TYPE_CHECKING, Literal, Optional, overload

from competitive_verifier import log

if TYPE_CHECKING:
    from _typeshed import StrOrBytesPath

    _StrOrListStr = str | list[str]


logger = getLogger(__name__)


@overload
def exec_command(
    command: "_StrOrListStr",
    *,
    text: Literal[False] = False,
    check: bool = False,
    capture_output: bool = False,
    env: dict[str, str] | None = None,
    cwd: Optional["StrOrBytesPath"] = None,
    group_log: bool = False,
) -> subprocess.CompletedProcess[bytes]: ...


@overload
def exec_command(
    command: "_StrOrListStr",
    *,
    text: Literal[True],
    check: bool = False,
    capture_output: bool = False,
    env: dict[str, str] | None = None,
    cwd: Optional["StrOrBytesPath"] = None,
    group_log: bool = False,
) -> subprocess.CompletedProcess[str]: ...


@overload
def exec_command(
    command: "_StrOrListStr",
    *,
    text: bool = False,
    check: bool = False,
    capture_output: bool = False,
    env: dict[str, str] | None = None,
    cwd: Optional["StrOrBytesPath"] = None,
    group_log: bool = False,
) -> subprocess.CompletedProcess[str] | subprocess.CompletedProcess[bytes]: ...


def exec_command(
    command: "_StrOrListStr",
    *,
    text: bool = False,
    check: bool = False,
    capture_output: bool = False,
    env: dict[str, str] | None = None,
    cwd: Optional["StrOrBytesPath"] = None,
    group_log: bool = False,
) -> subprocess.CompletedProcess[str] | subprocess.CompletedProcess[bytes]:
    encoding = sys.stdout.encoding if text else None
    if env:
        env = os.environ | env

    with (
        log.group(f"subprocess.run: {command}")
        if group_log
        else nullcontext(logger.info("subprocess.run: %s", command))
    ):
        return subprocess.run(
            command,
            shell=isinstance(command, str),
            text=text,
            check=check,
            env=env,
            cwd=cwd,
            capture_output=capture_output,
            encoding=encoding,
        )


@overload
def command_stdout(
    command: "_StrOrListStr",
    *,
    text: Literal[True] = True,
    check: bool = True,
    cwd: Optional["StrOrBytesPath"] = None,
) -> str: ...


@overload
def command_stdout(
    command: "_StrOrListStr",
    *,
    text: Literal[False],
    check: bool = True,
    cwd: Optional["StrOrBytesPath"] = None,
) -> bytes: ...


def command_stdout(
    command: "_StrOrListStr",
    *,
    text: bool = True,
    check: bool = True,
    cwd: Optional["StrOrBytesPath"] = None,
) -> str | bytes:
    return exec_command(
        command,
        text=text,
        cwd=cwd,
        check=check,
        capture_output=True,
    ).stdout



import logging
import pathlib
import uuid

import pytest
from pytest_mock import MockerFixture

from competitive_verifier.models import Problem
from competitive_verifier.models.problem import TestCaseFile as TstCaseFile
from competitive_verifier.oj.problem import (
    AOJArenaProblem,
    LocalProblem,
    YukicoderProblem,
)
from tests import LogComparer


@pytest.mark.allow_mkdir
def test_local_problem(caplog: pytest.LogCaptureFixture, testtemp: pathlib.Path):
    NOT_FOUND_RECORD = LogComparer("no cases found", logging.WARNING)
    DANGLING_OUTPUT_RECORD = LogComparer("dangling output case", logging.WARNING)

    problem = LocalProblem(testtemp)
    assert not problem.download_system_cases()
    assert caplog.records == [NOT_FOUND_RECORD]
    caplog.clear()
    assert list(problem.iter_system_cases()) == []
    assert caplog.records == [NOT_FOUND_RECORD]
    caplog.clear()

    (testtemp / "top.in").touch()
    (testtemp / "top.out").touch()
    (testtemp / "other.tmp").touch()

    (testtemp / "subdir").mkdir()
    (testtemp / "subdir" / "ss.in").touch()
    (testtemp / "subdir" / "ss.out").touch()
    (testtemp / "subdir" / "other.tmp").touch()

    assert problem.download_system_cases()
    assert list(problem.iter_system_cases()) == [
        TstCaseFile(
            name="subdir/ss",
            input_path=testtemp / "subdir/ss.in",
            output_path=testtemp / "subdir/ss.out",
        ),
        TstCaseFile(
            name="top",
            input_path=testtemp / "top.in",
            output_path=testtemp / "top.out",
        ),
    ]
    assert not caplog.records

    (testtemp / "subdir" / "only_in.in").touch()
    (testtemp / "subdir" / "only_out.out").touch()
    assert problem.download_system_cases()
    assert list(problem.iter_system_cases()) == [
        TstCaseFile(
            name="subdir/ss",
            input_path=testtemp / "subdir/ss.in",
            output_path=testtemp / "subdir/ss.out",
        ),
        TstCaseFile(
            name="top",
            input_path=testtemp / "top.in",
            output_path=testtemp / "top.out",
        ),
    ]
    assert caplog.records == [DANGLING_OUTPUT_RECORD, DANGLING_OUTPUT_RECORD]


test_problem_params = [
    (
        YukicoderProblem(problem_no=227),
        "68f6c4b6bd028f62cf05d0767b039e80",
        "YukicoderProblem.from_url('https://yukicoder.me/problems/no/227')",
    ),
    (
        YukicoderProblem(problem_id=227),
        "01d49ca1b113ce255a359b4263c99b15",
        "YukicoderProblem.from_url('https://yukicoder.me/problems/227')",
    ),
]


def test_problem_eq():
    yukicoder = YukicoderProblem(problem_id=227)

    assert yukicoder == YukicoderProblem(problem_id=227)
    assert yukicoder != YukicoderProblem(problem_no=227)
    assert yukicoder != AOJArenaProblem(arena_id="AOJ", alphabet="A")


@pytest.mark.parametrize(
    ("problem", "hash_id", "repr_str"),
    test_problem_params,
)
def test_problem(
    problem: Problem,
    hash_id: str,
    repr_str: str,
    mocker: MockerFixture,
    subtests: pytest.Subtests,
):
    cache_dir = pathlib.Path("/test_problem/dir") / uuid.uuid4().hex
    mocker.patch(
        "competitive_verifier.config.get_problem_cache_dir",
        return_value=cache_dir,
    )
    with subtests.test(msg="hash_id"):
        assert problem.hash_id == hash_id
    with subtests.test(msg="problem_directory"):
        assert problem.problem_directory == cache_dir / hash_id
    with subtests.test(msg="test_directory"):
        assert problem.test_directory == cache_dir / hash_id / "test"
    with subtests.test(msg="repr"):
        assert repr(problem) == repr_str



import pathlib
import sys
from contextlib import contextmanager
from dataclasses import asdict, dataclass
from logging import (
    DEBUG,
    ERROR,
    INFO,
    NOTSET,
    WARNING,
    Handler,
    LogRecord,
    basicConfig,
)
from typing import TextIO

import colorlog
from colorama import Fore, Style

from competitive_verifier import github


@dataclass
class GitHubMessageParams:
    title: str | None = None
    file: str | pathlib.Path | None = None
    col: int | None = None
    end_column: int | None = None
    line: int | None = None
    end_line: int | None = None


class GitHubActionsHandler(Handler):
    def __init__(
        self,
        *,
        stream: TextIO,
        level: int | str = NOTSET,
    ) -> None:
        super().__init__(level)
        self.stream = stream

    def emit(self, record: LogRecord) -> None:
        if record.levelno >= ERROR:
            command = "error"
        elif record.levelno >= WARNING:
            command = "warning"
        elif record.levelno >= INFO:
            command = "notice"
        else:
            if record.levelno >= DEBUG:
                other_name = (
                    ""
                    if record.name.startswith("competitive_verifier")
                    else f"{record.name}:{record.lineno}:"
                )
                github.debug(other_name + record.getMessage(), stream=self.stream)
            return

        if (gh := getattr(record, "github", None)) is None or not isinstance(
            gh, GitHubMessageParams
        ):
            return

        if not (gh.title or record.name.startswith("competitive_verifier")):
            gh.title = record.name

        github.message(command, record.getMessage(), stream=self.stream, **asdict(gh))


def configure_stderr_logging(
    default_level: int | None = None,
) -> None:  # pragma: no cover
    colorlog_handler = colorlog.StreamHandler(sys.stderr)
    colorlog_handler.setLevel(default_level or WARNING)
    colorlog_handler.setFormatter(
        colorlog.ColoredFormatter(
            "%(log_color)s%(levelname)s%(reset)s:%(name)s:%(lineno)d:%(message)s"
        )
    )

    handlers: list[Handler] = [colorlog_handler]

    if github.env.is_in_github_actions():
        handlers.append(GitHubActionsHandler(stream=sys.stderr))

    basicConfig(
        level=NOTSET,
        handlers=handlers,
    )


def _console_group(category: str, *, title: str, file: TextIO | None):
    print(
        (
            f"<------------- {Fore.CYAN}{category}:"
            f"{Fore.YELLOW}{title}"
            f"{Style.RESET_ALL} ------------->"
        ),
        file=file,
    )


@contextmanager
def group(title: str, *, stream: TextIO | None = None):
    if stream is None:
        stream = sys.stderr
    if github.env.is_in_github_actions():
        try:
            github.begin_group(title, stream=stream)
            yield
        finally:
            github.end_group(stream=stream)
    else:
        try:
            _console_group(" Start group", title=title, file=stream)
            yield
        finally:
            _console_group("Finish group", title=title, file=stream)



import datetime
import pathlib
import shutil
import subprocess
import sys
import tarfile
from typing import Any, Literal

import requests

from competitive_verifier import config
from competitive_verifier.models import FileResult, VerificationResult
from competitive_verifier.models import TestcaseResult as _TestcaseResult
from competitive_verifier.verify import verifier

from .utils import md5_number


class MockVerifyCommandResult(verifier.VerifyCommandResult):
    def model_dump_json(  # type: ignore[override]
        self,
        *,
        indent: Any = None,
        include: Any = None,
        exclude: Any = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        round_trip: bool = False,
        warnings: bool | Literal["none", "warn", "error"] = True,
    ) -> str:  # pragma: no cover
        return self.model_copy()._dump_super(  # noqa: SLF001
            indent=indent,
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            round_trip=round_trip,
            warnings=warnings,
        )

    def _dump_super(
        self,
        *,
        indent: Any = None,
        include: Any = None,
        exclude: Any = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        round_trip: bool = False,
        warnings: bool | Literal["none", "warn", "error"] = True,
    ) -> str:
        def rewriteVerifyCommandResult(result: verifier.VerifyCommandResult):
            result.total_seconds = len(result.files) * 1234.56 + 78
            result.files = {k: rewriteFileResult(k, v) for k, v in result.files.items()}
            return result

        def rewriteFileResult(path: pathlib.Path, file_result: FileResult):
            seed = path.as_posix().encode()
            file_result.verifications = [
                rewriteVerificationResult(seed, v) for v in file_result.verifications
            ]
            return file_result

        def rewriteVerificationResult(seed: bytes, verification: VerificationResult):
            seed += (verification.verification_name or "").encode()
            seed += verification.status.name.encode()

            verification.elapsed = md5_number(seed + b"elapsed") % 10000

            if verification.slowest:
                verification.slowest = verification.elapsed // 10
            if verification.heaviest:
                verification.heaviest = md5_number(seed + b"heaviest") % 1000

            if verification.testcases:
                verification.testcases = [
                    rewriteTestcaseResult(seed, c) for c in verification.testcases
                ]

            random_time = md5_number(seed + b"last_execution_time")
            verification.last_execution_time = datetime.datetime.fromtimestamp(
                random_time % 300000000000 / 100,
                tz=datetime.timezone(datetime.timedelta(hours=random_time % 25 - 12)),
            )

            return verification

        def rewriteTestcaseResult(seed: bytes, case: _TestcaseResult):
            seed += case.name.encode()
            seed += case.status.name.encode()

            case.elapsed = md5_number(seed + b"elapsed") % 1000 / 100
            case.memory = md5_number(seed + b"memory") % 10000 / 100
            return case

        rewriteVerifyCommandResult(self)
        return super().model_dump_json(
            indent=indent,
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            round_trip=round_trip,
            warnings=warnings,
        )


_library_checker_problems_tar_gz: bytes | None = None


def update_cloned_repository():  # pragma: no cover
    global _library_checker_problems_tar_gz  # noqa: PLW0603

    gz_path = config.get_cache_dir() / "library-checker-problems.tar.gz"
    repo_gz_path = config.get_cache_dir() / "repo.tar.gz"
    repo_path = config.get_cache_dir() / "library-checker-problems"
    if repo_path.is_dir():
        return

    if not gz_path.exists() and _library_checker_problems_tar_gz:
        gz_path.write_bytes(_library_checker_problems_tar_gz)
    if gz_path.exists():
        shutil.unpack_archive(gz_path, config.get_cache_dir())
        return

    config.get_cache_dir().mkdir(parents=True, exist_ok=True)
    res = requests.get(
        "https://github.com/yosupo06/library-checker-problems/archive/refs/heads/master.tar.gz",
        timeout=60,
    )

    repo_gz_path.write_bytes(res.content)
    shutil.unpack_archive(repo_gz_path, config.get_cache_dir())
    repo_gz_path.unlink(missing_ok=True)

    master_dir = config.get_cache_dir() / "library-checker-problems-master"

    gen_cmd = [sys.executable, "generate.py", "sample/aplusb/info.toml"]
    subprocess.check_call(gen_cmd, cwd=master_dir, timeout=60)

    with tarfile.open(gz_path, "w:gz") as gzp:
        gzp.add(master_dir, "library-checker-problems", filter=_match_aplusb)
    _library_checker_problems_tar_gz = gz_path.read_bytes()

    shutil.move(master_dir, repo_path)


def _match_aplusb(t: tarfile.TarInfo) -> tarfile.TarInfo | None:
    if (
        t.path.startswith("library-checker-problems/sample/aplusb")
        or (t.isdir() and "library-checker-problems/sample/aplusb".startswith(t.path))
        or t.isfile()
    ):
        return t
    return None



import contextlib
import os
import pathlib
import re
import subprocess
import tempfile
from dataclasses import dataclass
from functools import cache
from logging import getLogger
from typing import Protocol

logger = getLogger(__name__)


class GnuTimeRunner(Protocol):
    @property
    def gnu_time(self) -> str | None: ...

    def get_command(self, command: list[str]) -> list[str]: ...
    def get_memory(self) -> float | None:
        """Return the amount of memory used, in megabytes, if possible."""
        ...

    def clean(self) -> None: ...


class _GnuTimeRunnerDummy:
    @property
    def gnu_time(self) -> None:
        return None

    def get_command(self, command: list[str]) -> list[str]:
        return command

    def get_memory(self) -> float | None:
        pass

    def clean(self) -> None:
        pass


@dataclass
class _GnuTimeRunnerImpl:
    gnu_time: str
    tmpdir: tempfile.TemporaryDirectory[str]
    outfile: pathlib.Path

    def get_command(self, command: list[str]) -> list[str]:
        return [self.gnu_time, "-f", "%M", "-o", str(self.outfile), "--", *command]

    def get_memory(self) -> float | None:
        if self.outfile.exists() and (
            report := self.outfile.read_text("utf-8").strip()
        ):
            logger.debug("GNU time says: %s", report)
            tail = report.splitlines()[-1]
            if tail.isdigit():
                return int(tail) / 1000
        return None

    def clean(self) -> None:
        self.tmpdir.cleanup()


class GnuTimeWrapper(contextlib.AbstractContextManager["GnuTimeRunner"]):
    _gnu_time: str | None
    _runner: GnuTimeRunner

    def __init__(self, *, enabled: bool = True) -> None:
        super().__init__()
        self._gnu_time = time_command() if enabled else None
        self._runner = _GnuTimeRunnerDummy()

    def __enter__(self):
        if self._gnu_time:
            tmpdir = tempfile.TemporaryDirectory()
            self._runner = _GnuTimeRunnerImpl(
                gnu_time=self._gnu_time,
                tmpdir=tmpdir,
                outfile=pathlib.Path(tmpdir.name) / "gnu_time_report.txt",
            )
        return self._runner

    def __exit__(self, *excinfo: object):
        self._runner.clean()


@cache
def time_command() -> str | None:
    cmds = ["time", "gtime"]
    if os.name == "posix":
        cmds += ["/bin/time", "/usr/bin/time"]
    return _find_gnu_time(cmds)


def _find_gnu_time(gnu_time_candidate: list[str]) -> str | None:
    for gnu_time in gnu_time_candidate:
        if check_gnu_time(gnu_time):
            return gnu_time
    return None


def check_gnu_time(gnu_time: str) -> bool:
    try:
        with tempfile.TemporaryDirectory() as td:
            tmp = pathlib.Path(td) / "out"
            ret = subprocess.run(
                [
                    gnu_time,
                    "-f",
                    "%M KB",
                    "-o",
                    str(tmp),
                    "--",
                    "echo",
                    "check_gnu_time",
                ],
                encoding="utf-8",
                capture_output=True,
                check=True,
            )
            if (
                ret.returncode == 0
                and ret.stdout.rstrip() == "check_gnu_time"
                and re.match(r"^\d+ KB$", tmp.read_text("utf-8").strip())
            ):
                return True
    except NameError:
        raise  # NameError is not a runtime error caused by the environment, but a coding mistake
    except AttributeError:
        raise  # AttributeError is also a mistake
    except Exception:  # noqa: BLE001
        logger.debug("Failed to check gnu_time: %s", gnu_time, exc_info=True)
    return False



from .base import (
    Language,
    LanguageEnvironment,
    OjVerifyLanguageConfig,
    OjVerifyUserDefinedConfig,
)
from .cplusplus import (
    CPlusPlusLanguage,
    OjVerifyCPlusPlusConfig,
)
from .go import GoLanguage, OjVerifyGoConfig
from .haskell import (
    HaskellLanguage,
    OjVerifyHaskellConfig,
)
from .java import (
    JavaLanguage,
    OjVerifyJavaConfig,
)
from .nim import NimLanguage, OjVerifyNimConfig
from .python import PythonLanguage
from .ruby import (
    OjVerifyRubyConfig,
    RubyLanguage,
)
from .rust import (
    OjVerifyRustConfig,
    OjVerifyRustListDependenciesBackend,
    RustLanguage,
)
from .user_defined import UserDefinedLanguage

__all__ = [
    "CPlusPlusLanguage",
    "GoLanguage",
    "HaskellLanguage",
    "JavaLanguage",
    "Language",
    "LanguageEnvironment",
    "NimLanguage",
    "OjVerifyCPlusPlusConfig",
    "OjVerifyGoConfig",
    "OjVerifyHaskellConfig",
    "OjVerifyJavaConfig",
    "OjVerifyLanguageConfig",
    "OjVerifyNimConfig",
    "OjVerifyRubyConfig",
    "OjVerifyRustConfig",
    "OjVerifyRustListDependenciesBackend",
    "OjVerifyUserDefinedConfig",
    "PythonLanguage",
    "RubyLanguage",
    "RustLanguage",
    "UserDefinedLanguage",
]



import abc
import enum
import functools
import itertools
import json
import pathlib
import shutil
from collections import defaultdict
from collections.abc import Sequence
from enum import Enum
from logging import getLogger
from typing import Any, Literal

from pydantic import BaseModel, Field

from competitive_verifier.exec import command_stdout
from competitive_verifier.models import ShellCommand
from competitive_verifier.util import read_text_normalized

from .base import Language, LanguageEnvironment, OjVerifyLanguageConfig

# ruff: noqa: PLR2004

logger = getLogger(__name__)

_metadata_by_manifest_path: dict[pathlib.Path, dict[str, Any]] = {}
_cargo_checked_workspaces: set[pathlib.Path] = set()
_related_source_files_by_workspace: dict[
    pathlib.Path, dict[pathlib.Path, frozenset[pathlib.Path]]
] = {}


class OjVerifyRustListDependenciesBackend(BaseModel):
    kind: Literal["none", "cargo-udeps"]
    toolchain: str | None = None


class OjVerifyRustConfig(OjVerifyLanguageConfig):
    list_dependencies_backend: OjVerifyRustListDependenciesBackend | None = None


class _ListDependenciesBackend(abc.ABC, BaseModel):
    @abc.abstractmethod
    def list_dependencies(
        self, path: pathlib.Path, *, basedir: pathlib.Path
    ) -> list[pathlib.Path]: ...


class _NoBackend(_ListDependenciesBackend):
    def list_dependencies(
        self, path: pathlib.Path, *, basedir: pathlib.Path
    ) -> list[pathlib.Path]:
        return _list_dependencies_by_crate(
            path, basedir=basedir, cargo_udeps_toolchain=None
        )


class _CargoUdeps(_ListDependenciesBackend):
    toolchain: str = "nightly"

    def list_dependencies(
        self, path: pathlib.Path, *, basedir: pathlib.Path
    ) -> list[pathlib.Path]:
        return _list_dependencies_by_crate(
            path, basedir=basedir, cargo_udeps_toolchain=self.toolchain
        )


@functools.cache
def _list_dependencies_by_crate(
    path: pathlib.Path, *, basedir: pathlib.Path, cargo_udeps_toolchain: str | None
) -> list[pathlib.Path]:
    """The `list_dependencies` implementation for `_NoBackend` and `CargoUdeps`.

    Args:
        path (pathlib.Path): A main source file path of a target
        basedir (pathlib.Path): A parameter from `Language.list_dependencies`
        cargo_udeps_toolchain (str | None): If not `None`, use `cargo-udeps` with the specified toolchain to detect unused dependencies
    Returns:
        list[pathlib.Path]: A list of dependent `.rs` file paths
    Raises:
        RuntimeError: If any cargo command fails
    """
    path = basedir / path

    # We regard that a generated file does not depend on any files.
    for parent in path.parents:
        if (parent.parent / "Cargo.toml").exists() and parent.parts[-1] == "target":
            logger.warning("This is a generated file!: %s", path)
            return [path]

    metadata = _cargo_metadata(cwd=path.parent)

    # First, collects source files in the same crate.
    common_result = set(
        _source_files_in_same_targets(path, _related_source_files(basedir, metadata))
    )

    main_package_and_target = _find_target(metadata, path)
    if not main_package_and_target:
        return sorted(common_result)
    main_package, main_target = main_package_and_target

    packages_by_id = {p["id"]: p for p in metadata["packages"]}

    class DependencyNamespace(Enum):
        NORMAL_DEVELOPMENT = enum.auto()
        BUILD = enum.auto()

        @classmethod
        def from_dep_kind(cls, kind: str):
            if kind == "build":
                return cls.BUILD
            return cls.NORMAL_DEVELOPMENT

    # Collect the `(|dev-|build-)dependencies` into a <is a `build-dependency`> → (<"extern crate name"> → <package>) dictionary.
    dependencies: defaultdict[DependencyNamespace, dict[str, dict[str, Any]]] = (
        defaultdict(dict)
    )
    for dep in next(
        n["deps"] for n in metadata["resolve"]["nodes"] if n["id"] == main_package["id"]
    ):
        if _need_dev_deps(main_target) or any(
            k["kind"] is None for k in dep["dep_kinds"]
        ):
            dependencies[DependencyNamespace.NORMAL_DEVELOPMENT][dep["name"]] = (
                packages_by_id[dep["pkg"]]
            )
        if any(k["kind"] == "build" for k in dep["dep_kinds"]):
            dependencies[DependencyNamespace.BUILD][dep["name"]] = packages_by_id[
                dep["pkg"]
            ]

    # If `cargo_udeps_toolchain` is present, collects packages that are "unused" by `target`.
    unused_packages: defaultdict[DependencyNamespace, set[Any]] = defaultdict(set)
    if cargo_udeps_toolchain is not None:
        explicit_names_in_toml = {
            (DependencyNamespace.from_dep_kind(d["kind"]), d["rename"])
            for d in main_package["dependencies"]
            if d["rename"]
        }
        if not shutil.which("cargo-udeps"):
            raise RuntimeError("`cargo-udeps` not in $PATH")
        args: list[str] = [
            "rustup",
            "run",
            cargo_udeps_toolchain,
            "cargo",
            "udeps",
            "--output",
            "json",
            "--manifest-path",
            main_package["manifest_path"],
            *_target_option(main_target),
        ]
        unused_deps = json.loads(
            command_stdout(args, cwd=metadata["workspace_root"], check=False)
        )["unused_deps"].values()
        unused_dep = next(
            (
                u
                for u in unused_deps
                if u["manifest_path"] == main_package["manifest_path"]
            ),
            None,
        )
        if unused_dep:
            names_in_toml: list[tuple[DependencyNamespace, Any]] = [
                (DependencyNamespace.NORMAL_DEVELOPMENT, name_in_toml)
                for name_in_toml in [*unused_dep["normal"], *unused_dep["development"]]
            ]
            names_in_toml.extend(
                (DependencyNamespace.BUILD, name_in_toml)
                for name_in_toml in unused_dep["build"]
            )
            for dependency_namespace, name_in_toml in names_in_toml:
                if (dependency_namespace, name_in_toml) in explicit_names_in_toml:
                    # If the `name_in_toml` is explicitly renamed one, it equals to the `extern_crate_name`.
                    unused_package: Any = dependencies[dependency_namespace][
                        name_in_toml
                    ]["id"]
                else:
                    # Otherwise, it equals to the `package.name`.
                    unused_package = next(
                        p["id"]
                        for p in dependencies[dependency_namespace].values()
                        if p["name"] == name_in_toml
                    )
                unused_packages[dependency_namespace].add(unused_package)

    # Finally, adds source files related to the depended crates except:
    #
    # - those detected by cargo-udeps
    # - those come from Crates.io or Git repositories (e.g. `proconio`, other people's libraries including `ac-library-rs`)

    # `main_package` should always be included.
    # Note that cargo-udeps does not detect it if it is unused.
    # https://github.com/est31/cargo-udeps/pull/35
    depended_packages = [main_package]
    for dependency_namespace, values in dependencies.items():
        depended_packages.extend(
            depended_package
            for depended_package in values.values()
            if (
                depended_package["id"] not in unused_packages[dependency_namespace]
                and not depended_package["source"]
            )
        )

    ret = common_result

    for depended_package in depended_packages:
        depended_targets = [
            t
            for t in depended_package["targets"]
            if t != main_target and (_is_build(t) or _is_lib_or_proc_macro(t))
        ]
        assert len(depended_targets) <= 2
        for depended_target in depended_targets:
            related_source_files = _related_source_files(
                basedir,
                _cargo_metadata_by_manifest_path(
                    pathlib.Path(depended_package["manifest_path"])
                ),
            )
            ret |= _source_files_in_same_targets(
                pathlib.Path(depended_target["src_path"]).resolve(strict=True),
                related_source_files,
            )
    return sorted(ret)


def _related_source_files(
    basedir: pathlib.Path, metadata: dict[str, Any]
) -> dict[pathlib.Path, frozenset[pathlib.Path]]:
    """Collects all of the `.rs` files recognized by a workspace.

    Args:
        basedir (pathlib.Path): A parameter from `Language.list_dependencies`
        metadata (dict[str, Any]): "metadata" for a Cargo.toml file in the workspace
    Returns:
        dict[pathlib.Path, frozenset[pathlib.Path]]: A (main source file) → (other related files) map
    Raises:
        RuntimeError: If any cargo command fails
    """
    if pathlib.Path(metadata["workspace_root"]) in _related_source_files_by_workspace:
        return _related_source_files_by_workspace[
            pathlib.Path(metadata["workspace_root"])
        ]

    # Runs `cargo check` to generate `$target_directory/debug/deps/*.d`.
    if pathlib.Path(metadata["workspace_root"]) not in _cargo_checked_workspaces:
        command_stdout(
            [
                "cargo",
                "check",
                "--manifest-path",
                str(pathlib.Path(metadata["workspace_root"], "Cargo.toml")),
                "--workspace",
                "--all-targets",
            ],
            cwd=metadata["workspace_root"],
        )
        _cargo_checked_workspaces.add(pathlib.Path(metadata["workspace_root"]))

    ret: dict[pathlib.Path, frozenset[pathlib.Path]] = {}

    targets_in_workspace = itertools.chain.from_iterable(
        p["targets"]
        for p in metadata["packages"]
        if p["id"] in metadata["workspace_members"]
    )
    for target in targets_in_workspace:
        # Finds the **latest** "dep-info" file that contains a line in the following format, and parses the line.
        #
        # ```
        # <relative/absolute path to the `.d` file itself>: <relative/absolute path to the root source file> <relative/aboslute paths to the other related files>...
        # ```
        #
        # - https://github.com/rust-lang/cargo/blob/rust-1.49.0/src/cargo/core/compiler/fingerprint.rs#L1979-L1997
        # - https://github.com/rust-lang/cargo/blob/rust-1.49.0/src/cargo/core/compiler/fingerprint.rs#L1824-L1830
        if _is_build(target):
            dep_info_paths = pathlib.Path(
                metadata["target_directory"], "debug", "build"
            ).rglob(f"{_crate_name(target)}-*.d")
        elif _is_example(target):
            dep_info_paths = pathlib.Path(
                metadata["target_directory"], "debug", "examples"
            ).glob(f"{_crate_name(target)}-*.d")
        else:
            dep_info_paths = pathlib.Path(
                metadata["target_directory"], "debug", "deps"
            ).glob(f"{_crate_name(target)}-*.d")
        for dep_info_path in sorted(
            dep_info_paths, key=lambda p: p.stat().st_mtime_ns, reverse=True
        ):
            dep_info = read_text_normalized(dep_info_path)
            for line in dep_info.splitlines():
                ss = line.split(": ")
                if (
                    len(ss) == 2
                    and pathlib.Path(metadata["workspace_root"], ss[0]) == dep_info_path
                ):
                    paths: list[pathlib.Path] = []
                    it = iter(ss[1].split())
                    for s in it:
                        ss = s
                        while ss.endswith("\\"):
                            ss = ss.rstrip("\\") + " " + next(it)
                        path = pathlib.Path(metadata["workspace_root"], s).resolve(
                            strict=True
                        )
                        # Ignores paths that don't start with the `basedir`. (e.g. `/dev/null`, `/usr/local/share/foo/bar`)
                        if path.is_relative_to(basedir):
                            paths.append(path)
                    if paths[:1] == [
                        pathlib.Path(target["src_path"]).resolve(strict=True)
                    ]:
                        ret[paths[0]] = frozenset(paths[1:])
                        break
            else:
                continue
            break
        else:
            logger.error("no `.d` file for `%s`", target["name"])

    _related_source_files_by_workspace[pathlib.Path(metadata["workspace_root"])] = ret
    return ret


def _source_files_in_same_targets(
    path: pathlib.Path,
    related_source_files: dict[pathlib.Path, frozenset[pathlib.Path]],
) -> frozenset[pathlib.Path]:
    """Returns `.rs` file paths relating to `path`.

    Args:
        path (pathlib.Path): A main source file path of a target
        related_source_files (dict[pathlib.Path, frozenset[pathlib.Path]]): A (main source file) → (other related files) map
    Returns:
        frozenset[pathlib.Path]: A set of `.rs` file paths relating to `path`
    Raises:
        RuntimeError: If `path` is not found in `related_source_files` and is not related to any other files.
    """
    # If `p` is `src_path` of a target, it does not belong to any other target unless it's weirdly symlinked,
    if path in related_source_files:
        return frozenset({path, *related_source_files[path]})

    # Otherwise, it may be used by multiple targets with `#[path = ".."] mod foo;` or something.
    return frozenset(
        itertools.chain.from_iterable(
            {k, *v} for (k, v) in related_source_files.items() if path in v
        )
    ) or frozenset({path})


class RustLanguageEnvironment(LanguageEnvironment):
    @property
    def name(self) -> str:
        return "Rust"

    def get_compile_command(
        self, path: pathlib.Path, *, basedir: pathlib.Path, tempdir: pathlib.Path
    ) -> ShellCommand:
        path = basedir / path
        metadata = _cargo_metadata(cwd=path.parent)
        target = _ensure_target(metadata, path)
        return ShellCommand(
            command=["cargo", "build", "--release", *_target_option(target)],
            cwd=path.parent,
        )

    def get_execute_command(
        self, path: pathlib.Path, *, basedir: pathlib.Path, tempdir: pathlib.Path
    ) -> str:
        path = basedir / path
        metadata = _cargo_metadata(cwd=path.parent)
        target = _ensure_target(metadata, path)
        return str(
            pathlib.Path(
                metadata["target_directory"],
                "release",
                *([] if _is_bin(target) else ["examples"]),
                target["name"],
            )
        )


class RustLanguage(Language):
    config: OjVerifyRustConfig = Field(default_factory=OjVerifyRustConfig)

    @functools.cached_property
    def _list_dependencies_backend(self) -> _ListDependenciesBackend:
        list_dependencies_backend = self.config.list_dependencies_backend
        if (
            list_dependencies_backend is None
            or list_dependencies_backend.kind == "none"
        ):
            return _NoBackend()
        if list_dependencies_backend.kind == "cargo-udeps":
            if toolchain := list_dependencies_backend.toolchain:
                return _CargoUdeps(toolchain=toolchain)
            return _CargoUdeps()
        raise RuntimeError(
            "expected 'none' or 'cargo-udeps' for `languages.rust.list_dependencies_backend.kind`"
        )

    def list_dependencies(
        self, path: pathlib.Path, *, basedir: pathlib.Path
    ) -> list[pathlib.Path]:
        return self._list_dependencies_backend.list_dependencies(path, basedir=basedir)

    def list_environments(
        self, path: pathlib.Path, *, basedir: pathlib.Path
    ) -> Sequence[RustLanguageEnvironment]:
        return [RustLanguageEnvironment()]


def _cargo_metadata(cwd: pathlib.Path) -> dict[str, Any]:
    """Returns "metadata" for a Cargo.toml file in `cwd` or its parent directories.

    Args:
        cwd (pathlib.Path): The current working directory
    Returns:
        dict[str, Any]: Output of `cargo metadata` command
    Raises:
        ValueError: If `cwd` is not absolute or contains `..`
        RuntimeError: If no `Cargo.toml` is found
    """
    if not cwd.is_absolute() or ".." in cwd.parts:
        raise ValueError(
            f"the `cwd` parameter must be absolute and must not contain `..`: {cwd}"
        )

    # https://docs.rs/cargo/0.49.0/src/cargo/util/important_paths.rs.html#6-20
    for directory in [cwd, *cwd.parents]:
        manifest_path = directory / "Cargo.toml"
        if manifest_path.exists():
            return _cargo_metadata_by_manifest_path(manifest_path)
    raise RuntimeError(
        f"could not find `Cargo.toml` in `{cwd}` or any parent directory"
    )


def _cargo_metadata_by_manifest_path(manifest_path: pathlib.Path) -> dict[str, Any]:
    """Returns "metadata" for a certain `Cargo.toml`.

    Args:
        manifest_path (pathlib.Path): Path to a `Cargo.toml`
    Returns:
        dict[str, Any]: Output of `cargo metadata` command
    Raises:
        RuntimeError: If the `cargo metadata` command fails
    """
    if manifest_path in _metadata_by_manifest_path:
        return _metadata_by_manifest_path[manifest_path]

    metadata = _run_cargo_metadata(manifest_path)
    root_manifest_path = pathlib.Path(metadata["workspace_root"], "Cargo.toml")
    if root_manifest_path != manifest_path:
        metadata = _run_cargo_metadata(root_manifest_path)

    for key in [
        root_manifest_path,
        *(
            pathlib.Path(p["manifest_path"])
            for p in metadata["packages"]
            if p["id"] in metadata["workspace_members"]
        ),
    ]:
        _metadata_by_manifest_path[key] = metadata

    return metadata


def _run_cargo_metadata(manifest_path: pathlib.Path) -> dict[str, Any]:
    """Runs `cargo metadata` for a certain `Cargo.toml`.

    This function is considered to be executed just once for every Cargo.toml in the repository.
    For detailed information about `cargo metadata`, see:

    - <https://doc.rust-lang.org/cargo/commands/cargo-metadata.html#output-format>
    - <https://docs.rs/cargo_metadata>

    Args:
        manifest_path (pathlib.Path): Path to a `Cargo.toml`
    Returns:
        dict[str, Any]: Output of `cargo metadata` command
    Raises:
        RuntimeError: If the `cargo metadata` command fails
    """
    return json.loads(
        command_stdout(
            [
                "cargo",
                "metadata",
                "--format-version",
                "1",
                "--manifest-path",
                str(manifest_path),
            ],
            cwd=manifest_path.parent,
        )
    )


def _find_target(
    metadata: dict[str, Any],
    src_path: pathlib.Path,
) -> tuple[dict[str, Any], dict[str, Any]] | None:
    for package in metadata["packages"]:
        for target in package["targets"]:
            # A `src_path` may contain `..`
            # The path may not actually exist by being excluded from the package.
            if pathlib.Path(target["src_path"]).resolve() == src_path:
                return package, target
    return None


def _ensure_target(metadata: dict[str, Any], src_path: pathlib.Path) -> dict[str, Any]:
    package_and_target = _find_target(metadata, src_path)
    if not package_and_target:
        raise RuntimeError(f"{src_path} is not a main source file of any target")
    _, target = package_and_target
    return target


def _crate_name(target: dict[str, Any]) -> bool:
    return target["name"].replace("-", "_")


def _is_build(target: dict[str, Any]) -> bool:
    return target["kind"] == ["custom-build"]


def _is_lib_or_proc_macro(target: dict[str, Any]) -> bool:
    return target["kind"] in [["lib"], ["proc-macro"]]


def _is_bin(target: dict[str, Any]) -> bool:
    return target["kind"] == ["bin"]


def _is_example(target: dict[str, Any]) -> bool:
    return target["kind"] == ["example"]


def _need_dev_deps(target: dict[str, Any]) -> bool:
    # Comes from https://docs.rs/cargo/0.49.0/cargo/ops/enum.CompileFilter.html#method.need_dev_deps
    return not (_is_lib_or_proc_macro(target) or _is_bin(target))


def _target_option(target: dict[str, Any]) -> list[str]:
    if target["kind"] == ["bin"]:
        return ["--bin", target["name"]]
    if target["kind"] == ["example"]:
        return ["--example", target["name"]]
    if target["kind"] == ["test"]:
        return ["--test", target["name"]]
    if target["kind"] == ["bench"]:
        return ["--bench", target["name"]]
    return ["--lib"]



import pathlib
from typing import Literal, TextIO


def debug(message: str, *, stream: TextIO):
    print("::debug::" + message.replace("\n", "\\n"), file=stream)


def message(
    command: Literal["error", "warning", "notice"],
    message: str,
    *,
    title: str | None = None,
    file: str | pathlib.Path | None = None,
    col: int | None = None,
    end_column: int | None = None,
    line: int | None = None,
    end_line: int | None = None,
    stream: TextIO | None = None,
) -> None:
    """Print Github Actions style message.

    https://docs.github.com/en/actions/using-workflows/workflow-commands-for-github-actions

    Args:
        command: Command name (e.g., "error", "warning", "debug")
        message: Message body
    Keyword Args:
        title: Custom title
        file: Filename
        col: Column number, starting at 1
        end_column: End column number
        line: Line number, starting at 1
        end_line: End line number
        stream: Output stream
    """
    annotation = ",".join(
        f"{name}={value}"
        for name, value in (
            ("title", title),
            ("file", file.absolute() if isinstance(file, pathlib.Path) else file),
            ("col", col),
            ("endColumn", end_column),
            ("line", line),
            ("endLine", end_line),
        )
        if value is not None
    )

    message = message.replace("\n", "\\n")
    print(f"::{command} {annotation}::{message}", file=stream)


def begin_group(title: str, *, stream: TextIO | None = None):
    print(f"::group::{title}", file=stream)


def end_group(*, stream: TextIO | None = None):
    print("::endgroup::", file=stream)



import logging

import pytest
from pytest_mock import MockerFixture

from competitive_verifier import oj
from competitive_verifier.log import GitHubMessageParams
from competitive_verifier.oj.problem import NotLoggedInError, YukicoderProblem
from tests import LogComparer


def test_oj_download_not_supported(caplog: pytest.LogCaptureFixture):
    caplog.set_level(logging.NOTSET)
    assert oj.download("https://example.com/notfound") is False
    assert caplog.records == [
        LogComparer(
            'The URL "https://example.com/notfound" is not supported',
            logging.ERROR,
            github=GitHubMessageParams(),
        ),
    ]


def test_oj_download_not_logged_in(
    mocker: MockerFixture,
    caplog: pytest.LogCaptureFixture,
):
    caplog.set_level(logging.NOTSET)
    mocker.patch.object(
        YukicoderProblem,
        "download_system_cases",
        side_effect=NotLoggedInError("test error"),
    )

    assert oj.download("https://yukicoder.me/problems/no/499") is False
    assert caplog.records == [
        LogComparer(
            "download[Run]: https://yukicoder.me/problems/no/499",
            logging.INFO,
        ),
        LogComparer(
            "Login is required to download the problem. "
            "'https://yukicoder.me/problems/no/499'",
            logging.ERROR,
            github=GitHubMessageParams(),
        ),
    ]


def test_oj_download_error(
    mocker: MockerFixture,
    caplog: pytest.LogCaptureFixture,
):
    caplog.set_level(logging.NOTSET)
    mocker.patch.object(
        YukicoderProblem,
        "download_system_cases",
        side_effect=RuntimeError("test error"),
    )

    assert oj.download("https://yukicoder.me/problems/no/499") is False
    assert caplog.records == [
        LogComparer(
            "download[Run]: https://yukicoder.me/problems/no/499",
            logging.INFO,
        ),
        LogComparer(
            "Failed to download. 'https://yukicoder.me/problems/no/499'",
            logging.ERROR,
            github=GitHubMessageParams(),
        ),
    ]



import pathlib

from charset_normalizer import from_bytes


def to_relative(path: pathlib.Path) -> pathlib.Path | None:
    try:
        return path.resolve().relative_to(pathlib.Path.cwd().resolve())
    except ValueError:
        return None


def _resolve_referenced_path_inner(
    target_path: str,
    *,
    basedir: pathlib.Path,
):
    if target_path.startswith(("./", "../")):
        # a relative path
        path = basedir / pathlib.Path(target_path)
        if path.exists():
            return path
    elif target_path.startswith("//"):
        # from the document root
        path = pathlib.Path(target_path[2:])
        if path.exists():
            return path

    path = pathlib.Path(target_path)
    if path.exists():
        return path

    path = basedir / pathlib.Path(target_path)
    if path.exists():
        return path
    return None


def resolve_referenced_path(
    target_path: str,
    *,
    basedir: pathlib.Path,
) -> pathlib.Path | None:
    path = _resolve_referenced_path_inner(target_path, basedir=basedir)
    if path:
        path = path.resolve()
        if path.is_relative_to(pathlib.Path.cwd()):
            return path.relative_to(pathlib.Path.cwd())
    return None


def read_text_normalized(path: pathlib.Path) -> str:
    return normalize_bytes_text(path.read_bytes())


def normalize_bytes_text(b: bytes) -> str:
    return str(from_bytes(b).best())



import os
import pathlib
import tempfile
from collections.abc import Generator
from contextlib import nullcontext

import pytest
from pytest_mock import MockerFixture


@pytest.fixture
def mock_perf_counter(mocker: MockerFixture, request: pytest.FixtureRequest):
    if values := getattr(request, "param", None):
        assert isinstance(request.param, list)
    else:
        values = [float(i) for i in range(1000)]

    return mocker.patch("time.perf_counter", side_effect=values)


@pytest.fixture
def mockenv(mocker: MockerFixture, request: pytest.FixtureRequest):
    mocker.patch.dict(os.environ, request.param or {}, clear=True)


@pytest.fixture(autouse=True)
def prohibit_mkdir(mocker: MockerFixture, request: pytest.FixtureRequest):
    if "allow_mkdir" in request.keywords:
        return None
    return mocker.patch(
        "pathlib.Path.mkdir",
        side_effect=RuntimeError("mkdir is not allowed"),
    )


class TempContext(nullcontext[pathlib.Path]):
    def __init__(self, enter_result: pathlib.Path) -> None:
        super().__init__(enter_result)
        os.mkdir(enter_result)  # noqa: PTH102

    @property
    def name(self):
        return self.enter_result

    def cleanup(self): ...


@pytest.fixture
def testtemp(
    mocker: MockerFixture, monkeypatch: pytest.MonkeyPatch
) -> Generator[pathlib.Path, None, None]:
    with tempfile.TemporaryDirectory() as d:
        monkeypatch.chdir(d)
        tmp = pathlib.Path(d).resolve()

        def temp_dir():
            count = 0
            while True:
                count += 1
                yield TempContext(tmp / str(count))

        mocker.patch("tempfile.TemporaryDirectory", side_effect=temp_dir())
        yield tmp
        monkeypatch.undo()



import logging
import os
import sys
from io import StringIO

import pytest
from pytest_mock import MockerFixture

from competitive_verifier.log import GitHubActionsHandler, GitHubMessageParams, group

test_github_actions_handler_params = []


def test_github_actions_handler(capsys: pytest.CaptureFixture[str]):
    handler = GitHubActionsHandler(stream=sys.stderr)

    def getLogger(name: str):
        logger = logging.getLogger(name)
        logger.addHandler(handler)
        logger.setLevel(1)
        return logger

    GitHub = "GitHub"
    logger = getLogger("competitive_verifier.handler")

    logger.log(1, "Super low", extra={"github": GitHubMessageParams()})
    logger.debug("DebugLog")
    logger.debug("%sDebugLog\nFin", GitHub, extra={"github": GitHubMessageParams()})
    out, err = capsys.readouterr()
    assert out == ""
    assert err == "::debug::DebugLog\n::debug::GitHubDebugLog\\nFin\n"

    logger.info("InfoLog")
    logger.info("%sInfoLog\nFin", GitHub, extra={"github": GitHubMessageParams()})
    out, err = capsys.readouterr()
    assert out == ""
    assert err == "::notice ::GitHubInfoLog\\nFin\n"

    logger.warning("WarningLog")
    logger.warning("%sWarningLog\nFin", GitHub, extra={"github": GitHubMessageParams()})
    out, err = capsys.readouterr()
    assert out == ""
    assert err == "::warning ::GitHubWarningLog\\nFin\n"

    logger.exception("ErrorLog")
    logger.error(
        "%sErrorLog\nFin",
        GitHub,
        extra={
            "github": GitHubMessageParams(
                title="TestTitle",
                file="foo.txt",
                col=10,
                end_column=20,
                line=100,
                end_line=101,
            )
        },
    )
    out, err = capsys.readouterr()
    assert out == ""
    assert (
        err
        == "::error title=TestTitle,file=foo.txt,col=10,endColumn=20,line=100,endLine=101::GitHubErrorLog\\nFin\n"
    )

    logger2 = getLogger("otherlib.handler")
    logger2.error("%sErrorLog\nFin", GitHub, extra={"github": GitHubMessageParams()})
    out, err = capsys.readouterr()
    assert out == ""
    assert err == "::error title=otherlib.handler::GitHubErrorLog\\nFin\n"

    logger2 = getLogger("otherlib.handler")
    logger2.error("%sErrorLog\nFin", GitHub, extra={"github": GitHubMessageParams()})
    out, err = capsys.readouterr()
    assert out == ""
    assert err == "::error title=otherlib.handler::GitHubErrorLog\\nFin\n"


class TestLogGroup:
    def test_log_group_github_actions(
        self, mocker: MockerFixture, capsys: pytest.CaptureFixture[str]
    ):
        mocker.patch.dict(os.environ, {"GITHUB_ACTIONS": "true"}, clear=True)
        with group("TestTitle"):
            out, err = capsys.readouterr()
            assert out == ""
            assert err == "::group::TestTitle\n"

        out, err = capsys.readouterr()
        assert out == ""
        assert err == "::endgroup::\n"

    def test_log_group_github_actions_stream(self, mocker: MockerFixture):
        mocker.patch.dict(os.environ, {"GITHUB_ACTIONS": "true"}, clear=True)
        with StringIO() as fp:
            with group("TestTitle", stream=fp):
                fp.seek(0)
                assert fp.read() == "::group::TestTitle\n"
                fp.seek(0)
                fp.truncate(0)
            fp.seek(0)
            assert fp.read() == "::endgroup::\n"

    def test_log_group_console(self, capsys: pytest.CaptureFixture[str]):
        with group("TestTitle"):
            out, err = capsys.readouterr()
            assert out == ""
            assert (
                err
                == "<------------- \x1b[36m Start group:\x1b[33mTestTitle\x1b[0m ------------->\n"
            )

        out, err = capsys.readouterr()
        assert out == ""
        assert (
            err
            == "<------------- \x1b[36mFinish group:\x1b[33mTestTitle\x1b[0m ------------->\n"
        )

    def test_log_group_console_stream(self):
        with StringIO() as fp:
            with group("TestTitle", stream=fp):
                fp.seek(0)
                assert (
                    fp.read()
                    == "<------------- \x1b[36m Start group:\x1b[33mTestTitle\x1b[0m ------------->\n"
                )
                fp.seek(0)
                fp.truncate(0)
            fp.seek(0)
            assert (
                fp.read()
                == "<------------- \x1b[36mFinish group:\x1b[33mTestTitle\x1b[0m ------------->\n"
            )



from .oj_download import main as download
from .oj_test import main as test
from .problem import LocalProblem, problem_from_url
from .resolver import OjResolve

__all__ = [
    "LocalProblem",
    "OjResolve",
    "download",
    "problem_from_url",
    "test",
]



import datetime
import logging
import os
import pathlib
from typing import Any

import pytest
from pytest_mock import MockerFixture

from competitive_verifier.log import GitHubMessageParams
from competitive_verifier.models import (
    ConstVerification,
    FileResult,
    ProblemVerification,
    ResultStatus,
    VerificationInput,
    VerificationResult,
    VerifyCommandResult,
)
from competitive_verifier.verify.verifier import BaseVerifier, SplitState
from tests import LogComparer

SUCCESS = ResultStatus.SUCCESS
FAILURE = ResultStatus.FAILURE


class NotSkippableConstVerification(ConstVerification):
    @property
    def is_lightweight(self) -> bool:
        return False


class MockVerifier(BaseVerifier):
    def __init__(
        self,
        varifications: Any = None,
        *,
        verification_time: datetime.datetime,
        prev_result: VerifyCommandResult | None = None,
        split_state: SplitState | None = None,
    ) -> None:
        super().__init__(
            verifications=VerificationInput.model_validate(varifications),
            verification_time=verification_time,
            prev_result=prev_result,
            split_state=split_state,
            default_tle=10,
            default_mle=256,
            timeout=10,
        )

    def get_file_timestamp(self, path: pathlib.Path) -> datetime.datetime:
        return datetime.datetime(2005, 1, 2, 15, 4, 5)


test_verify_params: list[tuple[MockVerifier, dict[str, Any]]] = [
    (
        MockVerifier(
            {
                "files": {
                    "lib/hoge1.py": {},
                    "test/foo.py": {
                        "dependencies": ["lib/hoge1.py"],
                        "verification": [
                            NotSkippableConstVerification(status=ResultStatus.SUCCESS)
                        ],
                    },
                    "test/skip.py": {
                        "dependencies": ["lib/hoge1.py"],
                        "verification": [
                            ConstVerification(status=ResultStatus.SUCCESS)
                        ],
                    },
                }
            },
            verification_time=datetime.datetime(2007, 1, 2, 15, 4, 5),
        ),
        {
            "total_seconds": 8.0,
            "files": {
                "test/foo.py": FileResult(
                    newest=True,
                    verifications=[
                        VerificationResult(
                            status=ResultStatus.SUCCESS,
                            elapsed=2.0,
                            last_execution_time=datetime.datetime(2007, 1, 2, 15, 4, 5),
                        ),
                    ],
                ),
                "test/skip.py": FileResult(
                    newest=True,
                    verifications=[
                        VerificationResult(
                            status=ResultStatus.SUCCESS,
                            elapsed=2.0,
                            last_execution_time=datetime.datetime(2007, 1, 2, 15, 4, 5),
                        ),
                    ],
                ),
            },
        },
    ),
    (
        MockVerifier(
            {
                "files": {
                    "lib/hoge1.py": {},
                    "test/foo.py": {
                        "dependencies": ["lib/hoge1.py"],
                        "verification": [
                            NotSkippableConstVerification(status=ResultStatus.SUCCESS)
                        ],
                    },
                    "test/foo1.py": {
                        "dependencies": ["lib/hoge1.py"],
                        "verification": [
                            NotSkippableConstVerification(status=ResultStatus.SUCCESS)
                        ],
                    },
                    "test/foo2.py": {
                        "dependencies": ["lib/hoge1.py"],
                        "verification": [
                            NotSkippableConstVerification(status=ResultStatus.SUCCESS)
                        ],
                    },
                    "test/foo3.py": {
                        "dependencies": ["lib/hoge1.py"],
                        "verification": [
                            NotSkippableConstVerification(status=ResultStatus.SUCCESS)
                        ],
                    },
                    "test/skip.py": {
                        "dependencies": ["lib/hoge1.py"],
                        "verification": [
                            ConstVerification(status=ResultStatus.SUCCESS)
                        ],
                    },
                }
            },
            verification_time=datetime.datetime(2007, 1, 2, 15, 4, 5),
            split_state=SplitState(size=2, index=0),
        ),
        {
            "total_seconds": 12.0,
            "files": {
                "test/foo.py": FileResult(
                    newest=True,
                    verifications=[
                        VerificationResult(
                            status=ResultStatus.SUCCESS,
                            elapsed=2.0,
                            last_execution_time=datetime.datetime(2007, 1, 2, 15, 4, 5),
                        ),
                    ],
                ),
                "test/foo1.py": FileResult(
                    newest=True,
                    verifications=[
                        VerificationResult(
                            status=ResultStatus.SUCCESS,
                            elapsed=2.0,
                            last_execution_time=datetime.datetime(2007, 1, 2, 15, 4, 5),
                        ),
                    ],
                ),
                "test/skip.py": FileResult(
                    newest=True,
                    verifications=[
                        VerificationResult(
                            status=ResultStatus.SUCCESS,
                            elapsed=2.0,
                            last_execution_time=datetime.datetime(2007, 1, 2, 15, 4, 5),
                        ),
                    ],
                ),
            },
        },
    ),
    (
        MockVerifier(
            {
                "files": {
                    "lib/hoge1.py": {},
                    "test/foo.py": {
                        "dependencies": ["lib/hoge1.py"],
                        "verification": [
                            NotSkippableConstVerification(status=ResultStatus.SUCCESS)
                        ],
                    },
                    "test/foo1.py": {
                        "dependencies": ["lib/hoge1.py"],
                        "verification": [
                            NotSkippableConstVerification(status=ResultStatus.SUCCESS)
                        ],
                    },
                    "test/foo2.py": {
                        "dependencies": ["lib/hoge1.py"],
                        "verification": [
                            NotSkippableConstVerification(status=ResultStatus.SUCCESS)
                        ],
                    },
                    "test/foo3.py": {
                        "dependencies": ["lib/hoge1.py"],
                        "verification": [
                            NotSkippableConstVerification(status=ResultStatus.SUCCESS)
                        ],
                    },
                    "test/skip.py": {
                        "dependencies": ["lib/hoge1.py"],
                        "verification": [
                            ConstVerification(status=ResultStatus.SUCCESS)
                        ],
                    },
                }
            },
            verification_time=datetime.datetime(2007, 1, 2, 15, 4, 5),
            split_state=SplitState(size=2, index=1),
        ),
        {
            "total_seconds": 9.0,
            "files": {
                "test/foo2.py": FileResult(
                    newest=True,
                    verifications=[
                        VerificationResult(
                            status=ResultStatus.SUCCESS,
                            elapsed=2.0,
                            last_execution_time=datetime.datetime(2007, 1, 2, 15, 4, 5),
                        ),
                    ],
                ),
                "test/foo3.py": FileResult(
                    newest=True,
                    verifications=[
                        VerificationResult(
                            status=ResultStatus.SUCCESS,
                            elapsed=2.0,
                            last_execution_time=datetime.datetime(2007, 1, 2, 15, 4, 5),
                        ),
                    ],
                ),
            },
        },
    ),
    (
        MockVerifier(
            {
                "files": {
                    "lib/hoge1.py": {},
                    "test/foo.py": {
                        "dependencies": ["lib/hoge1.py"],
                        "verification": [
                            NotSkippableConstVerification(status=ResultStatus.SUCCESS)
                        ],
                    },
                    "test/foo1.py": {
                        "dependencies": ["lib/hoge1.py"],
                        "verification": [
                            NotSkippableConstVerification(status=ResultStatus.SUCCESS)
                        ],
                    },
                    "test/foo2.py": {
                        "dependencies": ["lib/hoge1.py"],
                        "verification": [
                            NotSkippableConstVerification(status=ResultStatus.SUCCESS)
                        ],
                    },
                    "test/foo3.py": {
                        "dependencies": ["lib/hoge1.py"],
                        "verification": [
                            NotSkippableConstVerification(status=ResultStatus.SUCCESS)
                        ],
                    },
                    "test/skip.py": {
                        "dependencies": ["lib/hoge1.py"],
                        "verification": [
                            ConstVerification(status=ResultStatus.SUCCESS)
                        ],
                    },
                }
            },
            prev_result=VerifyCommandResult.model_validate(
                {
                    "total_seconds": 6.0,
                    "files": {
                        "test/foo.py": FileResult(
                            newest=True,
                            verifications=[
                                VerificationResult(
                                    status=ResultStatus.SUCCESS,
                                    elapsed=1.0,
                                    last_execution_time=datetime.datetime(
                                        2005, 12, 2, 15, 4, 5
                                    ),
                                ),
                            ],
                        ),
                        "test/foo1.py": FileResult(
                            newest=True,
                            verifications=[
                                VerificationResult(
                                    status=ResultStatus.SUCCESS,
                                    elapsed=1.0,
                                    last_execution_time=datetime.datetime(
                                        2005, 12, 2, 15, 4, 5
                                    ),
                                ),
                            ],
                        ),
                        "test/foo2.py": FileResult(
                            newest=True,
                            verifications=[
                                VerificationResult(
                                    status=ResultStatus.SUCCESS,
                                    elapsed=1.0,
                                    last_execution_time=datetime.datetime(
                                        2005, 12, 2, 15, 4, 5
                                    ),
                                ),
                            ],
                        ),
                        "test/foo3.py": FileResult(
                            newest=True,
                            verifications=[
                                VerificationResult(
                                    status=ResultStatus.SUCCESS,
                                    elapsed=1.0,
                                    last_execution_time=datetime.datetime(
                                        2000, 1, 2, 15, 4, 5
                                    ),
                                ),
                            ],
                        ),
                        "test/skip.py": FileResult(
                            newest=False,
                            verifications=[
                                VerificationResult(
                                    status=ResultStatus.SUCCESS,
                                    elapsed=1.0,
                                    last_execution_time=datetime.datetime(
                                        2000, 1, 2, 15, 4, 5
                                    ),
                                ),
                            ],
                        ),
                    },
                }
            ),
            verification_time=datetime.datetime(2007, 1, 2, 15, 4, 5),
        ),
        {
            "total_seconds": 8.0,
            "files": {
                "test/foo.py": FileResult(
                    newest=False,
                    verifications=[
                        VerificationResult(
                            status=ResultStatus.SUCCESS,
                            elapsed=1.0,
                            last_execution_time=datetime.datetime(
                                2005, 12, 2, 15, 4, 5
                            ),
                        ),
                    ],
                ),
                "test/foo1.py": FileResult(
                    newest=False,
                    verifications=[
                        VerificationResult(
                            status=ResultStatus.SUCCESS,
                            elapsed=1.0,
                            last_execution_time=datetime.datetime(
                                2005, 12, 2, 15, 4, 5
                            ),
                        ),
                    ],
                ),
                "test/foo2.py": FileResult(
                    newest=False,
                    verifications=[
                        VerificationResult(
                            status=ResultStatus.SUCCESS,
                            elapsed=1.0,
                            last_execution_time=datetime.datetime(
                                2005, 12, 2, 15, 4, 5
                            ),
                        ),
                    ],
                ),
                "test/foo3.py": FileResult(
                    newest=True,
                    verifications=[
                        VerificationResult(
                            status=ResultStatus.SUCCESS,
                            elapsed=2.0,
                            last_execution_time=datetime.datetime(2007, 1, 2, 15, 4, 5),
                        ),
                    ],
                ),
                "test/skip.py": FileResult(
                    newest=True,
                    verifications=[
                        VerificationResult(
                            status=ResultStatus.SUCCESS,
                            elapsed=2.0,
                            last_execution_time=datetime.datetime(2007, 1, 2, 15, 4, 5),
                        ),
                    ],
                ),
            },
        },
    ),
]


@pytest.mark.usefixtures("mock_perf_counter")
@pytest.mark.parametrize(
    ("verifier", "expected"),
    test_verify_params,
)
def test_verify(
    verifier: MockVerifier,
    expected: Any,
    mocker: MockerFixture,
):
    mocker.patch.object(pathlib.Path, "exists", return_value=True)
    assert verifier.verify(download=False) == VerifyCommandResult.model_validate(
        expected
    )


test_verify_timeout_params: list[
    tuple[
        MockVerifier,
        list[float],
        dict[str, Any],
    ]
] = [
    (
        MockVerifier(
            {
                "files": {
                    "lib/hoge1.py": {},
                    "test/foo.py": {
                        "dependencies": ["lib/hoge1.py"],
                        "verification": [
                            NotSkippableConstVerification(status=ResultStatus.SUCCESS)
                        ],
                    },
                }
            },
            verification_time=datetime.datetime(2007, 1, 2, 15, 4, 5),
        ),
        [0.0, 11.0, 12.0, 13.0, 14.0],
        {
            "total_seconds": 14.0,
            "files": {
                "test/foo.py": FileResult(
                    newest=True,
                    verifications=[
                        VerificationResult(
                            status=ResultStatus.SKIPPED,
                            elapsed=1.0,
                            last_execution_time=datetime.datetime(2007, 1, 2, 15, 4, 5),
                        ),
                    ],
                ),
            },
        },
    ),
    (
        MockVerifier(
            {
                "files": {
                    "lib/hoge1.py": {},
                    "test/foo.py": {
                        "dependencies": ["lib/hoge1.py"],
                        "verification": [
                            NotSkippableConstVerification(status=ResultStatus.SUCCESS)
                        ],
                    },
                }
            },
            verification_time=datetime.datetime(2007, 1, 2, 15, 4, 5),
        ),
        [0.0, 5.0, 11.0, 12.0, 13.0],
        {
            "total_seconds": 13.0,
            "files": {
                "test/foo.py": FileResult(
                    newest=True,
                    verifications=[
                        VerificationResult(
                            status=ResultStatus.SKIPPED,
                            elapsed=1.0,
                            last_execution_time=datetime.datetime(2007, 1, 2, 15, 4, 5),
                        ),
                    ],
                ),
            },
        },
    ),
    (
        MockVerifier(
            {
                "files": {
                    "lib/hoge1.py": {},
                    "test/foo.py": {
                        "dependencies": ["lib/hoge1.py"],
                        "verification": [
                            NotSkippableConstVerification(status=ResultStatus.SUCCESS)
                        ],
                    },
                }
            },
            verification_time=datetime.datetime(2007, 1, 2, 15, 4, 5),
        ),
        [0.0, 5.0, 6.0, 11.0, 12.0, 13.0],
        {
            "total_seconds": 13.0,
            "files": {
                "test/foo.py": FileResult(
                    newest=True,
                    verifications=[
                        VerificationResult(
                            status=ResultStatus.SKIPPED,
                            elapsed=6.0,
                            last_execution_time=datetime.datetime(2007, 1, 2, 15, 4, 5),
                        ),
                    ],
                ),
            },
        },
    ),
    (
        MockVerifier(
            {
                "files": {
                    "lib/hoge1.py": {},
                    "test/foo1.py": {
                        "dependencies": ["lib/hoge1.py"],
                        "verification": [
                            NotSkippableConstVerification(status=ResultStatus.SUCCESS)
                        ],
                    },
                    "test/foo2.py": {
                        "dependencies": ["lib/hoge1.py"],
                        "verification": [
                            NotSkippableConstVerification(status=ResultStatus.SUCCESS)
                        ],
                    },
                }
            },
            verification_time=datetime.datetime(2007, 1, 2, 15, 4, 5),
        ),
        [0.0, 5.0, 6.0, 7.0, 8.0, 11.0, 12.0, 13.0, 14.0],
        {
            "total_seconds": 14.0,
            "files": {
                "test/foo1.py": FileResult(
                    newest=True,
                    verifications=[
                        VerificationResult(
                            status=ResultStatus.SUCCESS,
                            elapsed=2.0,
                            last_execution_time=datetime.datetime(2007, 1, 2, 15, 4, 5),
                        ),
                    ],
                ),
                "test/foo2.py": FileResult(
                    newest=True,
                    verifications=[
                        VerificationResult(
                            status=ResultStatus.SKIPPED,
                            elapsed=1.0,
                            last_execution_time=datetime.datetime(2007, 1, 2, 15, 4, 5),
                        ),
                    ],
                ),
            },
        },
    ),
    (
        MockVerifier(
            {
                "files": {
                    "lib/hoge1.py": {},
                    "test/foo.py": {
                        "dependencies": ["lib/hoge1.py"],
                        "verification": [
                            NotSkippableConstVerification(status=ResultStatus.SUCCESS),
                            NotSkippableConstVerification(status=ResultStatus.SUCCESS),
                        ],
                    },
                }
            },
            verification_time=datetime.datetime(2007, 1, 2, 15, 4, 5),
        ),
        [0.0, 5.0, 6.0, 7.0, 8.0, 11.0, 12.0, 13.0],
        {
            "total_seconds": 13.0,
            "files": {
                "test/foo.py": FileResult(
                    newest=True,
                    verifications=[
                        VerificationResult(
                            status=ResultStatus.SUCCESS,
                            elapsed=2.0,
                            last_execution_time=datetime.datetime(2007, 1, 2, 15, 4, 5),
                        ),
                        VerificationResult(
                            status=ResultStatus.SKIPPED,
                            elapsed=1.0,
                            last_execution_time=datetime.datetime(2007, 1, 2, 15, 4, 5),
                        ),
                    ],
                ),
            },
        },
    ),
]


@pytest.mark.usefixtures("mock_perf_counter")
@pytest.mark.parametrize(
    ("verifier", "mock_perf_counter", "expected"),
    test_verify_timeout_params,
    indirect=["mock_perf_counter"],
)
def test_verify_timeout(
    mocker: MockerFixture,
    verifier: MockVerifier,
    expected: dict[str, Any],
):
    """Test timeout exception scenarios in enumerate_verifications."""
    mocker.patch.object(pathlib.Path, "exists", return_value=True)
    download = mocker.patch(
        "competitive_verifier.verify.verifier.run_download", return_value=True
    )

    result = verifier.verify(download=False)
    assert result == VerifyCommandResult.model_validate(expected)
    download.assert_not_called()


@pytest.mark.usefixtures("mock_perf_counter")
def test_verify_download_error(
    mocker: MockerFixture,
    caplog: pytest.LogCaptureFixture,
):
    mocker.patch("competitive_verifier.oj.download", return_value=False)
    verification = [
        ProblemVerification(
            name="foo",
            command="false",
            problem="https://judge.yosupo.jp/problem/aplusb",
        ),
        ProblemVerification(
            name="bar",
            command="false",
            problem="https://judge.yosupo.jp/problem/aplusb",
        ),
    ]
    verifier = MockVerifier(
        {
            "files": {
                "lib/hoge1.py": {},
                "test/foo.py": {
                    "dependencies": ["lib/hoge1.py"],
                    "verification": verification,
                },
            }
        },
        verification_time=datetime.datetime(2007, 1, 2, 15, 4, 5),
    )
    result = verifier.verify(download=True)
    assert result.model_dump(exclude_none=True) == {
        "total_seconds": 4.0,
        "files": {
            pathlib.Path("test/foo.py"): {
                "newest": True,
                "verifications": [
                    {
                        "elapsed": 1.0,
                        "last_execution_time": datetime.datetime(2007, 1, 2, 15, 4, 5),
                        "status": ResultStatus.FAILURE,
                    },
                ],
            }
        },
    }
    assert caplog.records == [
        LogComparer(
            f"Failed to download: {verification}",
            logging.ERROR,
            github=GitHubMessageParams(),
        ),
    ]


@pytest.mark.usefixtures("mock_perf_counter")
@pytest.mark.parametrize("is_github_actions", [False, True])
def test_verify_compile_error(
    is_github_actions: bool,
    mocker: MockerFixture,
    caplog: pytest.LogCaptureFixture,
    capsys: pytest.CaptureFixture[str],
):
    mocker.patch.dict(os.environ, {"GITHUB_ACTIONS": str(is_github_actions)})

    mocker.patch.object(
        pathlib.Path, "resolve", return_value=pathlib.Path("/any/dir/test/mock.py")
    )
    mocker.patch(
        "competitive_verifier.models.ProblemVerification.run_compile_command",
        return_value=False,
    )
    verifier = MockVerifier(
        {
            "files": {
                "lib/hoge1.py": {},
                "test/foo.py": {
                    "dependencies": ["lib/hoge1.py"],
                    "verification": [
                        ProblemVerification(
                            name="foo",
                            command="false",
                            problem="https://judge.yosupo.jp/problem/aplusb",
                        ),
                        ProblemVerification(
                            name="bar",
                            command="false",
                            problem="https://judge.yosupo.jp/problem/aplusb",
                        ),
                    ],
                },
            }
        },
        verification_time=datetime.datetime(2007, 1, 2, 15, 4, 5),
    )
    result = verifier.verify(download=False)
    assert result.model_dump(exclude_none=True) == {
        "total_seconds": 6.0,
        "files": {
            pathlib.Path("test/foo.py"): {
                "newest": True,
                "verifications": [
                    {
                        "verification_name": "foo",
                        "elapsed": 1.0,
                        "last_execution_time": datetime.datetime(2007, 1, 2, 15, 4, 5),
                        "status": ResultStatus.FAILURE,
                    },
                    {
                        "verification_name": "bar",
                        "elapsed": 1.0,
                        "last_execution_time": datetime.datetime(2007, 1, 2, 15, 4, 5),
                        "status": ResultStatus.FAILURE,
                    },
                ],
            }
        },
    }
    assert caplog.records == [
        LogComparer(
            f"Failed to compile: {pathlib.Path('test/foo.py')}, "
            'verification={"name":"foo","command":"false","problem":"https://judge.yosupo.jp/problem/aplusb"}',
            logging.ERROR,
            github=GitHubMessageParams(file=pathlib.Path("test/foo.py")),
        ),
        LogComparer(
            f"Failed to compile: {pathlib.Path('test/foo.py')}, verification="
            '{"name":"bar","command":"false","problem":"https://judge.yosupo.jp/problem/aplusb"}',
            logging.ERROR,
            github=GitHubMessageParams(file=pathlib.Path("test/foo.py")),
        ),
    ]

    out, err = capsys.readouterr()

    assert out == ""
    if is_github_actions:
        assert err == (
            "::group::current_verification_files\n::endgroup::\n"
            "::group::Verify: test/foo.py\n::endgroup::\n"
        )
    else:
        assert err == (
            "<------------- \x1b[36m Start group:\x1b[33mcurrent_verification_files\x1b[0m ------------->\n"
            "<------------- \x1b[36mFinish group:\x1b[33mcurrent_verification_files\x1b[0m ------------->\n"
            "<------------- \x1b[36m Start group:\x1b[33mVerify: test/foo.py\x1b[0m ------------->\n"
            "<------------- \x1b[36mFinish group:\x1b[33mVerify: test/foo.py\x1b[0m ------------->\n"
        )


@pytest.mark.usefixtures("mock_perf_counter")
def test_verify_error(
    caplog: pytest.LogCaptureFixture,
    capsys: pytest.CaptureFixture[str],
):
    class ErrorVerification(ConstVerification):
        @property
        def is_lightweight(self) -> bool:
            return False

        def run(self, *args: Any, **kwargs: Any):  # pyright: ignore[reportIncompatibleMethodOverride]
            raise RuntimeError("ErrorVerification")

    verifier = MockVerifier(
        {
            "files": {
                "lib/hoge1.py": {},
                "test/foo.py": {
                    "dependencies": ["lib/hoge1.py"],
                    "verification": [
                        ErrorVerification(
                            name="foo",
                            status=ResultStatus.FAILURE,
                        ),
                    ],
                },
            }
        },
        verification_time=datetime.datetime(2007, 1, 2, 15, 4, 5),
    )
    result = verifier.verify(download=False)
    assert result.model_dump(exclude_none=True) == {
        "total_seconds": 5.0,
        "files": {
            pathlib.Path("test/foo.py"): {
                "newest": True,
                "verifications": [
                    {
                        "verification_name": "foo",
                        "elapsed": 2.0,
                        "last_execution_time": datetime.datetime(2007, 1, 2, 15, 4, 5),
                        "status": ResultStatus.FAILURE,
                    },
                ],
            }
        },
    }
    assert caplog.records == [
        LogComparer(
            f"Failed to verify: {pathlib.Path('test/foo.py')}, "
            "ErrorVerification(name='foo', type='const', status=<ResultStatus.FAILURE: 'failure'>)",
            logging.ERROR,
            github=GitHubMessageParams(),
        ),
    ]

    out, err = capsys.readouterr()

    assert out == ""
    assert err == (
        "<------------- \x1b[36m Start group:\x1b[33mcurrent_verification_files\x1b[0m ------------->\n"
        "<------------- \x1b[36mFinish group:\x1b[33mcurrent_verification_files\x1b[0m ------------->\n"
        "<------------- \x1b[36m Start group:\x1b[33mVerify: test/foo.py\x1b[0m ------------->\n"
        "<------------- \x1b[36mFinish group:\x1b[33mVerify: test/foo.py\x1b[0m ------------->\n"
    )


@pytest.mark.usefixtures("mock_perf_counter")
@pytest.mark.parametrize("is_github_actions", [False, True])
def test_verify_failure(
    is_github_actions: bool,
    mocker: MockerFixture,
    caplog: pytest.LogCaptureFixture,
    capsys: pytest.CaptureFixture[str],
):
    mocker.patch.dict(os.environ, {"GITHUB_ACTIONS": str(is_github_actions)})
    mocker.patch.object(
        pathlib.Path, "resolve", return_value=pathlib.Path("/any/dir/test/mock.py")
    )
    verifier = MockVerifier(
        {
            "files": {
                "lib/hoge1.py": {},
                "test/foo.py": {
                    "dependencies": ["lib/hoge1.py"],
                    "verification": [
                        ConstVerification(
                            name="foo",
                            status=ResultStatus.FAILURE,
                        ),
                    ],
                },
            }
        },
        verification_time=datetime.datetime(2007, 1, 2, 15, 4, 5),
    )
    result = verifier.verify(download=False)
    assert result.model_dump(exclude_none=True) == {
        "total_seconds": 4.0,
        "files": {
            pathlib.Path("test/foo.py"): {
                "newest": True,
                "verifications": [
                    {
                        "verification_name": "foo",
                        "elapsed": 2.0,
                        "last_execution_time": datetime.datetime(2007, 1, 2, 15, 4, 5),
                        "status": ResultStatus.FAILURE,
                    },
                ],
            }
        },
    }
    assert not caplog.records

    out, err = capsys.readouterr()

    if is_github_actions:
        assert out == ""
        assert err == "::group::current_verification_files\n::endgroup::\n"
    else:
        assert out == ""
        assert err == (
            "<------------- \x1b[36m Start group:\x1b[33mcurrent_verification_files\x1b[0m ------------->\n"
            "<------------- \x1b[36mFinish group:\x1b[33mcurrent_verification_files\x1b[0m ------------->\n"
        )


@pytest.mark.usefixtures("mock_perf_counter")
def test_failure_result():
    class ResultConstVerification(ConstVerification):
        def run(self, *args: Any, **kwargs: Any):  # pyright: ignore[reportIncompatibleMethodOverride]
            return VerificationResult(
                verification_name="mockresult",
                status=self.status,
                elapsed=1.2,
                last_execution_time=datetime.datetime(2007, 1, 2, 10, 4, 5),
            )

    verifier = MockVerifier(
        {
            "files": {
                "lib/hoge1.py": {},
                "test/foo.py": {
                    "dependencies": ["lib/hoge1.py"],
                    "verification": [
                        ResultConstVerification(
                            name="foo",
                            status=ResultStatus.FAILURE,
                        ),
                    ],
                },
            }
        },
        verification_time=datetime.datetime(2007, 1, 2, 15, 4, 5),
    )
    result = verifier.verify(download=False)
    assert result.model_dump(exclude_none=True) == {
        "total_seconds": 3.0,
        "files": {
            pathlib.Path("test/foo.py"): {
                "newest": True,
                "verifications": [
                    {
                        "verification_name": "mockresult",
                        "elapsed": 1.2,
                        "last_execution_time": datetime.datetime(2007, 1, 2, 10, 4, 5),
                        "status": ResultStatus.FAILURE,
                    },
                ],
            }
        },
    }



import datetime
import enum
from collections.abc import Sequence
from typing import Annotated, Any

from pydantic import BaseModel, ConfigDict, PlainSerializer, model_validator
from pydantic.alias_generators import to_camel

from competitive_verifier.models import ForcePosixPath, SortedPathList, TestcaseResult


class FileType(enum.Enum):
    LIBRARY = enum.auto()
    TEST = enum.auto()


class StatusIcon(str, enum.Enum):
    def __new__(cls, file_type: FileType, result: str, _: object) -> "StatusIcon":
        value = f"{file_type.name}_{result}"
        member = str.__new__(cls, value)
        member._value_ = value
        return member

    def __init__(self, file_type: FileType, result: str, is_success: bool) -> None:  # noqa: FBT001
        super().__init__()
        self.file_type = file_type
        self.result = result
        self.is_success = is_success

    LIBRARY_ALL_AC = (FileType.LIBRARY, "ALL_AC", True)
    LIBRARY_PARTIAL_AC = (FileType.LIBRARY, "PARTIAL_AC", True)
    LIBRARY_SOME_WA = (FileType.LIBRARY, "SOME_WA", False)
    LIBRARY_ALL_WA = (FileType.LIBRARY, "ALL_WA", False)
    LIBRARY_NO_TESTS = (FileType.LIBRARY, "NO_TESTS", True)
    TEST_ACCEPTED = (FileType.TEST, "ACCEPTED", True)
    TEST_WRONG_ANSWER = (FileType.TEST, "WRONG_ANSWER", False)
    TEST_WAITING_JUDGE = (FileType.TEST, "WAITING_JUDGE", True)

    @property
    def is_failed(self) -> bool:
        return not self.is_success

    @property
    def is_test(self) -> bool:
        return self.file_type == FileType.TEST


class RenderBaseModel(BaseModel):
    model_config = ConfigDict(
        alias_generator=to_camel,
        populate_by_name=True,
        extra="allow",
    )


class RenderLink(RenderBaseModel):
    path: ForcePosixPath
    filename: str
    icon: StatusIcon
    title: str | None = None

    @model_validator(mode="after")
    def validate_title(self: "RenderLink") -> "RenderLink":
        if self.path.as_posix() == self.title:
            self.title = None
        return self


class Dependency(RenderBaseModel):
    type: str
    files: list[RenderLink]


class EmbeddedCode(RenderBaseModel):
    name: str
    code: str


class EnvTestcaseResult(RenderBaseModel, TestcaseResult):
    environment: str | None


class CategorizedIndex(RenderBaseModel):
    name: str
    pages: list[RenderLink]


class IndexFiles(RenderBaseModel):
    type: str
    categories: list[CategorizedIndex]


class PageRenderData(RenderBaseModel):
    path: ForcePosixPath
    path_extension: str
    title: str | None
    embedded: list[EmbeddedCode]

    timestamp: Annotated[
        datetime.datetime,
        PlainSerializer(lambda x: str(x), return_type=str, when_used="json"),
    ]
    attributes: dict[str, Any]
    testcases: list[EnvTestcaseResult] | None = None

    is_failed: bool
    is_verification_file: bool
    verification_status: StatusIcon

    depends_on: SortedPathList
    required_by: SortedPathList
    verified_with: SortedPathList

    document_path: ForcePosixPath | None = None
    dependencies: list[Dependency]


class CodePageData(PageRenderData):
    document_content: str | None


class MultiCodePageData(RenderBaseModel):
    path: ForcePosixPath
    verification_status: StatusIcon
    is_failed: bool
    codes: Sequence[CodePageData]
    dependencies: list[Dependency]


class IndexRenderData(RenderBaseModel):
    top: list[IndexFiles]



import pathlib
from collections.abc import Sequence
from logging import getLogger
from typing import Any, Literal

from pydantic import Field, ValidationInfo, field_validator

from .base import LanguageEnvironment, OjVerifyUserDefinedConfig
from .user_defined import UserDefinedLanguage

logger = getLogger(__name__)


class OjVerifyJavaConfig(OjVerifyUserDefinedConfig):
    execute: None = None  # pyright: ignore[reportIncompatibleVariableOverride]

    @field_validator("execute", "compile", mode="before")
    @classmethod
    def name_must_contain_space(cls, v: Any, info: ValidationInfo) -> None:  # noqa: ANN401
        if v is None:
            return
        raise ValueError(f'You cannot overwrite "{info.field_name}" for Java language')


class JavaLanguageEnvironment(LanguageEnvironment):
    @property
    def name(self) -> str:
        return "Java"

    def get_compile_command(
        self, path: pathlib.Path, *, basedir: pathlib.Path, tempdir: pathlib.Path
    ) -> list[str]:
        return ["javac", str(basedir / path)]

    def get_execute_command(
        self, path: pathlib.Path, *, basedir: pathlib.Path, tempdir: pathlib.Path
    ) -> list[str]:
        relative_path = (basedir / path).relative_to(basedir)
        class_path = ".".join([*relative_path.parent.parts, relative_path.stem])
        return ["java", class_path]


class JavaLanguage(UserDefinedLanguage):
    extension: Literal["java"] = "java"  # pyright: ignore[reportIncompatibleVariableOverride]
    config: OjVerifyJavaConfig = Field(default_factory=OjVerifyJavaConfig)  # pyright: ignore[reportIncompatibleVariableOverride]

    def list_environments(
        self, path: pathlib.Path, *, basedir: pathlib.Path
    ) -> Sequence[LanguageEnvironment]:
        return [JavaLanguageEnvironment()]



from typing import Literal

from pydantic import Field

from competitive_verifier.models import ShellCommand, ShellCommandLike

from .base import OjVerifyUserDefinedConfig
from .user_defined import UserDefinedLanguage


class OjVerifyRubyConfig(OjVerifyUserDefinedConfig):
    execute: ShellCommandLike = Field(
        default_factory=lambda: ShellCommand(command=["ruby", "{basedir}/{path}"]),
    )


class RubyLanguage(UserDefinedLanguage):
    extension: Literal["rb"] = "rb"  # pyright: ignore[reportIncompatibleVariableOverride]
    config: OjVerifyRubyConfig = Field(default_factory=OjVerifyRubyConfig)  # pyright: ignore[reportIncompatibleVariableOverride]



import sys
from logging import getLogger

logger = getLogger(__name__)


def ulimit_stack() -> None:
    """Run `ulimit -s unlimited`."""
    if sys.platform not in ["win32", "darwin"]:  # pragma: no cover
        import resource  # noqa: PLC0415

        _, hard = resource.getrlimit(resource.RLIMIT_STACK)
        resource.setrlimit(resource.RLIMIT_STACK, (hard, hard))


def try_ulimit_stack() -> None:  # pragma: no cover
    """Run `ulimit -s unlimited` and ignore any errors."""
    try:
        ulimit_stack()
    except Exception:  # noqa: BLE001
        logger.warning("failed to increase the stack size[ulimit]")



import logging
import os
import sys
from typing import Any

import pytest
from pytest_mock import MockerFixture

from competitive_verifier.exec import exec_command
from tests import LogComparer


def test_exec_group_log(
    capsys: pytest.CaptureFixture[str],
    caplog: pytest.LogCaptureFixture,
    mocker: MockerFixture,
):
    def mockrun(command: Any, **kwargs: dict[str, Any]):
        print("mockrun", file=sys.stderr)  # noqa: T201

    mocker.patch.dict(os.environ, {"GITHUB_ACTIONS": "true"}, clear=True)
    mocker.patch("subprocess.run", side_effect=mockrun)

    exec_command("echo 1", group_log=True)

    assert not caplog.records
    out, err = capsys.readouterr()
    assert out == ""
    assert err == "::group::subprocess.run: echo 1\nmockrun\n::endgroup::\n"


def test_exec_not_group_log(
    capsys: pytest.CaptureFixture[str],
    caplog: pytest.LogCaptureFixture,
    mocker: MockerFixture,
):
    def mockrun(command: Any, **kwargs: dict[str, Any]):
        logging.getLogger("test").error("mockrun")

    mocker.patch.dict(os.environ, {"GITHUB_ACTIONS": "true"}, clear=True)
    mocker.patch("subprocess.run", side_effect=mockrun)
    caplog.set_level(logging.NOTSET)

    exec_command("echo 1")

    assert caplog.records == [
        LogComparer("subprocess.run: echo 1", logging.INFO),
        LogComparer("mockrun", logging.ERROR),
    ]
    out, err = capsys.readouterr()
    assert out == ""
    assert err == ""



import logging
import os
import pathlib
from abc import ABC, abstractmethod
from argparse import ArgumentParser
from logging import getLogger
from typing import TYPE_CHECKING, TypedDict, cast, get_args

from pydantic import BaseModel, Field

from competitive_verifier import github, summary

from .log import GitHubMessageParams, configure_stderr_logging

if TYPE_CHECKING:
    from competitive_verifier.models import VerifyCommandResult

COMPETITIVE_VERIFY_FILES_PATH = "COMPETITIVE_VERIFY_FILES_PATH"

logger = getLogger(__name__)


class _SubcommandInfo(TypedDict):
    name: str
    help: str | None


class BaseArguments(ABC, BaseModel):
    model_config = {"extra": "ignore"}

    @classmethod
    def get_subcommand_info(cls) -> _SubcommandInfo | None:
        f = cls.model_fields.get("subcommand")
        if f is None or (subcommand := get_args(f.annotation)[0]) is None:
            return None

        return {"name": cast("str", subcommand), "help": f.description}

    @abstractmethod
    def run(self) -> bool: ...

    @classmethod
    def add_parser(cls, parser: ArgumentParser):
        pass


class VerifyFilesJsonArgumentsMixin(BaseArguments):
    @classmethod
    def _required(cls) -> bool:
        return True

    @classmethod
    def add_parser(cls, parser: ArgumentParser):
        super().add_parser(parser)

        default = os.getenv(COMPETITIVE_VERIFY_FILES_PATH) if cls._required() else None
        parser.add_argument(
            "--verify-json",
            dest="verify_files_json",
            default=default or None,
            required=cls._required() and not bool(default),
            help="File path of verify_files.json. default: environ variable $COMPETITIVE_VERIFY_FILES_PATH",
            type=pathlib.Path,
        )


class VerifyFilesJsonArguments(VerifyFilesJsonArgumentsMixin):
    verify_files_json: pathlib.Path


class OptionalVerifyFilesJsonArguments(VerifyFilesJsonArgumentsMixin):
    verify_files_json: pathlib.Path | None

    @classmethod
    def _required(cls) -> bool:
        return False


class ResultJsonArguments(BaseArguments):
    result_json: list[pathlib.Path]

    @classmethod
    def add_parser(cls, parser: ArgumentParser):
        super().add_parser(parser)
        parser.add_argument(
            "result_json",
            nargs="+",
            help="Json files which is result of `verify`",
            type=pathlib.Path,
        )


class IgnoreErrorArguments(BaseArguments):
    ignore_error: bool = True

    @classmethod
    def add_parser(cls, parser: ArgumentParser):
        super().add_parser(parser)
        parser.add_argument(
            "--check-error",
            help="Exit not zero if failed verification exists",
            dest="ignore_error",
            action="store_false",
        )


class WriteSummaryArguments(BaseArguments):
    write_summary: bool = False

    @classmethod
    def add_parser(cls, parser: ArgumentParser):
        super().add_parser(parser)
        parser.add_argument(
            "--write-summary",
            action="store_true",
            help="Write GitHub Actions summary",
        )

    def write_result(self, result: "VerifyCommandResult"):
        if self.write_summary:
            gh_summary_path = github.env.get_step_summary_path()
            if gh_summary_path and gh_summary_path.parent.exists():
                with gh_summary_path.open("w", encoding="utf-8") as fp:
                    summary.write_summary(fp, result)
            else:
                logger.warning(
                    "write_summary=True but not found $GITHUB_STEP_SUMMARY",
                    extra={"github": GitHubMessageParams()},
                )


class IncludeExcludeArguments(BaseArguments):
    include: list[str] = Field(default_factory=list)
    exclude: list[str] = Field(default_factory=list)

    @classmethod
    def add_parser(cls, parser: ArgumentParser):
        super().add_parser(parser)
        parser.add_argument(
            "--include",
            nargs="*",
            help="Included file",
            default=[],
            type=str,
        )
        parser.add_argument(
            "--exclude",
            nargs="*",
            help="Excluded file",
            default=[],
            type=str,
        )


class VerboseArguments(BaseArguments):
    verbose: bool = False

    @abstractmethod
    def _run(self) -> bool: ...
    def run(self) -> bool:
        default_level = logging.INFO
        if self.verbose:
            default_level = logging.DEBUG
        configure_stderr_logging(default_level)
        return self._run()

    @classmethod
    def add_parser(cls, parser: ArgumentParser):
        super().add_parser(parser)
        parser.add_argument(
            "-v", "--verbose", action="store_true", help="Show debug level log."
        )
